from collections import namedtuple
from pathlib import Path
import logging

# import datatest
import dew_gwdata as gd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import ticker as mticker

# import pytest
from matplotlib import gridspec
import wrap_technote as tn

from scipy import optimize, stats

from .utils import *
from .gwutils import *

logger = get_logger()


def filter_tds_observations(df, return_removals=False, **kwargs):
    """Filter out salinity observations which are not suitable for analysis.

    Args:
        df (pandas DataFrame): observations
        return_removals (bool): return the removals which end up being applied.

    Returns:
        :class:`pandas.DataFrame` or tuple of two dataframes if
        *return_removals* is True.

    Other keyword arguments are passed to
    :func:`wrap_technote.load_qc_removals`.

    Removes rows where:

    1. anomalous_ind == "Y" AND
    2. measured_during == "D"


    """
    df = df[~df.collected_date.isnull()]
    df = df[~df.tds.isnull()]

    removals = [
        {
            "reason": "SA Geodata records with anomalous_ind != 'N'",
            "idx": df.anomalous_ind != "N",
        },
        {
            "reason": "SA Geodata records with measured_during in 'D', 'R', 'U'",
            "idx": df.measured_during.isin(["D"]),
        },
    ]
    removals += load_qc_removals(df, parameter="TDS", dt_col="collected_date", **kwargs)

    rdfs = []
    comb_idx = np.array([False for x in range(len(df))])
    for removal in removals:
        idx = removal["idx"].values
        rdf = pd.DataFrame(df[idx])
        rdf["reason"] = removal["reason"]
        rdfs.append(rdf)
        comb_idx = np.logical_or(comb_idx, idx)

    df = df[~comb_idx]

    if return_removals:
        return df, pd.concat(rdfs)
    else:
        return df


def reduce_to_annual_tds(df, reduction_func="mean"):
    """Reduce salinity data to an annual value.

    Args:
        df (pd.DataFrame): salinity data with columns
            "well_id", "collected_date" (datetimes), "tds"
        reduction_func (str, lambda): aggregation function
            for pandas.GroupBy.agg e.g. "mean", "median"

    Return:
        pandas.DataFrame: dataframe with columns "collected_year",
        "well_id", "easting", "northing", "latitude", "longitude", "tds"

    You can get the necessary data for this by for example from
    a salinity resource key `resource`:

    .. code-block:: python

        >>> df = resource.read_data("validated_data", "valid_data")

    """
    if len(df) == 0:
        return pd.DataFrame(
            columns=[
                "collected_year",
                "well_id",
                "easting",
                "northing",
                "latitude",
                "longitude",
                "tds",
            ]
        )
    else:
        tdsann = (
            df.groupby(
                [
                    "well_id",
                    "easting",
                    "northing",
                    "longitude",
                    "latitude",
                    df.collected_date.dt.year,
                ]
            )
            .tds.agg(reduction_func)
            .reset_index()
            .rename(columns={"collected_date": "collected_year"})
        )
        tdsann["decade"] = tdsann.collected_year.map(
            generate_yearly_periods(spacing=10)
        )
        tdsann["period_5yr"] = tdsann.collected_year.map(
            generate_yearly_periods(spacing=5)
        )

        return tdsann


def calculate_annual_tds_stats(df):
    """Return annual statistics by year.

    Args:
        df (pd.DataFrame): tds data - one row per well per year
            e.g. see :func:`reduce_to_annual_tds`

    Returns:
        :class:`pandas.DataFrame`: dataframe with year as index
        and columns "n_wells",
        "tds_mean", "tds_median", and "tds_stdev"

    """
    if len(df):
        grouped = df.groupby("collected_year")
        annstats = pd.concat(
            [
                grouped.well_id.count(),
                grouped.tds.mean(),
                grouped.tds.median(),
                grouped.tds.std(),
            ],
            axis=1,
        )
        annstats.columns = ["n_wells", "tds_mean", "tds_median", "tds_stdev"]
        return annstats
    else:
        return pd.DataFrame(columns=["n_wells", "tds_mean", "tds_median", "tds_stdev"])


def get_annual_mean_salinity_data(tdsann, well_id):
    """Return annual mean salinity data points.

    Args:
        tdsann (pd.DataFrame): columns "well_id", "collected_year" (int) and "tds" (float)
        well_id (str): value to group by

    Returns:
        pd.DataFrame: a dataframe with columns "well_id", "collected_date", and "tds",
        with the date set to June 30th of each year's value, containing only
        that well's data.

    Note that *tdsann* can be generated from an ordinary table `tds` of salinity data
    using code such as:

    .. code-block:: python

        >>> tdsann = (
        ...         tds.groupby([tds.well_id, tds.collected_date.dt.year])
        ...         .tds.mean()
        ...         .reset_index()
        ...         .rename(columns={"collected_date": "collected_year"})
        ...     )

    """
    idx = lambda x: x.well_id == well_id
    df = tdsann[idx].sort_values("collected_year")
    df["collected_date"] = [pd.Timestamp(y, 6, 30) for y in df.collected_year]
    return df


def linear_salinity_trend(
    df, tds_col="tds", dt_col="collected_date", regression_method="least-squares"
):
    """Calculate a linear salinity trend.

    Args:
        df (pandas.DataFrame): data to regress against.
        tds_col (str): column in `df` that contains salinity data in mg/L
        dt_col (str): column in `df` that contains pd.Timestamp datetimes.
        regression_method (str): regression_method (str): either "least-squares"
            or "L2" for ordinary least squares or "least-absolute-deviation" or
            "L1" for deviation

    Returns:
        :class:`pandas.DataFrame`: pandas dataframe with columns:

        - "observed" (coming from ``df[tds_col]``)
        - "ndays" (calculated from ``df[dt_col]``)
        - "timestamp" (coming from ``df[dt_col]``)
        - "predicted" (trend line value)
        - "start_value" (first observed value from period)
        - "end_value" (last observed value from period)
        - "slope_yr_ndays" (slope of the trend line in units of mg/L/day)
        - "y_int_ndays" (trend line intercept at day=0, almost totally meaningless)
        - "slope_yr" (trend line slope in units of mg/L/y).

    This function uses :func:`wrap_technote.linear_trend` to do the regression.

    """
    ndays = df[dt_col].apply(lambda dt: tn.date_to_decimal(dt)).values
    dfx = pd.DataFrame(
        {"observed": df[tds_col], "ndays": ndays, "timestamp": df[dt_col]}, index=None
    )
    tline = linear_trend(
        dfx, param_col="observed", dt_col="ndays", regression_method=regression_method
    ).to_dict()
    predicted = ndays * tline["slope_yr_ndays"] + tline["y_int_ndays"]
    dfx["start_value"] = df[tds_col].iloc[0]
    dfx["end_value"] = df[tds_col].iloc[-1]
    dfx["predicted"] = predicted
    for k, v in tline.items():
        dfx[k] = v
    return dfx


def calculate_historical_salinity_trends(
    df, trend_length_years=10, years="all", regression_method="least-squares"
):
    """Calculate salinity trends for each year possible.

    Args:
        df (pd.DataFrame): should have columns *collected_date* (datetime), *tds* (float)
        trend_length_years (int): number of years the trend should be calculated over.
        years (str or iterable): either "all", "final", or an iterable of integers.
            Trends will calculated as of each year, so for example if *df* has data
            from 1990 to 2020, if *years* is `"all"`, a trend line and status will
            be calculated as at 1991, 1992, and so on until 2021.

    Returns:
        tuple: with three items:

        - years (ndarray): the years for which trends have been calculated
        - dfs (list, each item a dict of DataFrames): one under the key "df" shows the
          original data with columns "well_id", "collected_year", "tds", "collected_date",
          "mean". Another under "trend_df" shows the trend line data against observed
          data. It has columns "observed", "ndays", "timestamp", "start_value",
          "end_value", "predicted", "slope_yr_ndays", "y_int_ndays", and "slope_yr".
          One set of dataframes is returned for each year in `years`, i.e. the length
          of `dfs` is the length of `years`.
        - results (pd.DataFrame): shows a summary for each trend period calculated.
          Columns are: "well_id", "year", "start_year", "current_tds", "current_pct_diff",
          "mean", "n_years_data", "slope_mgl_yr", "slope_pct_change_yr",

    If, for a given trend period (i.e. `year - (trend_length_years - 1)`), the number
    of datapoints is 2 or greater, then the trend period at that year will be included
    in the results (below). If not, it will be silently excluded. As a result, note that
    the returned `years` ndarray may not be contiguous, i.e. it may have missing years.

    """

    if years == "all":
        year0 = df.collected_date.dt.year.min() + 1
        year1 = df.collected_date.dt.year.max()
        years = [y for y in range(year0, year1 + 1, 1)]
    elif years == "final":
        years = [df.collected_date.dt.year.max()]
    logger.debug(f"YEARS: {years}")
    dfs = []
    table_results = []
    years_ran = []
    well_id = df.well_id.unique()[0]
    for year in years:
        logger.debug(f"calculating trend for {year}")
        if len(df[df.collected_date.dt.year == year]) == 0:
            logger.debug(f"  skipping, no data in {year}")
            continue
        dfy = df[df.collected_date.dt.year <= year].sort_values("collected_date")
        dfy = dfy.dropna(subset=["tds"])
        start_year = year - (trend_length_years - 1)
        tdfy = dfy[dfy.collected_date.dt.year >= start_year]
        logger.debug(
            f"well_id={well_id} trend @ year={year} - start_year={start_year} - len(tdfy)={len(tdfy)}"
        )
        if len(tdfy) == 0:
            logger.debug(
                f"  skipping, only one data point between {start_year} and {year}"
            )
            continue
        tdf = linear_salinity_trend(tdfy, regression_method=regression_method)

        # QUESTION!!!!!!!!! is the "mean" the 10 year mean, or the whole lot of data?
        # Probably needs to be much more than 10 years.

        mean = dfy.tds.mean()
        current_tds = dfy.tds.iloc[-1]
        slope_mgl_yr = tdf.slope_yr.iloc[0]
        dfy["mean"] = mean
        pct_diff = ((current_tds - mean) / mean) * 100
        diff_word = "above" if pct_diff > 0 else "below"
        label_text = f"(<= {year}): {abs(pct_diff):.0f}% {diff_word} average"
        slope_pct_change_yr = (slope_mgl_yr / tdf.iloc[-1].predicted) * 100
        dfs_result = {
            "df": dfy,
            "trend_df": tdf,
        }
        table_result = {
            "well_id": well_id,
            "year": year,
            "start_year": start_year,
            "n_years_data": len(tdfy),
            "current_tds": current_tds,
            "mean": mean,
            "current_pct_diff": pct_diff,
            "slope_mgl_yr": slope_mgl_yr,
            "slope_pct_change_yr": slope_pct_change_yr,
            "slope_pct_change": slope_pct_change_yr * trend_length_years,
            "label": label_text,
            "trend_length_years": trend_length_years,
        }
        logger.debug(f"Results: {table_result}")
        years_ran.append(year)
        dfs.append(dfs_result)
        table_results.append(table_result)
    return np.array(years_ran), dfs, pd.DataFrame(table_results)


def generate_salinity_bins(
    range_min,
    range_max,
    step,
    extrema=1e10,
    label_fmt="{bin_left}% to {bin_right}% {word}",
    end_label_fmt="More than {bin_left}% {word}",
    word_positive="increase",
    word_negative="decrease",
):
    """Generate bins/classes for salinity indicator values.

    .. todo:: Document this function

    Args:
        range_min (float)
        range_max (float)
        step (int)
        extrema (float)
        label_fmt (str)
        end_label (str)
        word_positive (str)
        word_negative (str)

    Returns:
        dict:

    """
    # bins = np.arange(range_min, range_max + step, step)
    bins = np.linspace(
        range_min, range_max, int(np.ceil((range_max - range_min) / step) + 1)
    )
    bins = np.concatenate([np.array([-1 * extrema]), bins, np.array([extrema])])
    bin_centres = []
    bin_labels = []
    bin_classes = []
    for i, value in enumerate(bins[:-1]):
        if value < 0:
            multiplier = -1
            word = word_negative
        else:
            multiplier = 1
            word = word_positive
        value_upper = bins[i + 1]
        bounds = sorted([abs(value), abs(value_upper)])
        label = label_fmt.format(
            word=word,
            bin_left="{:.0f}".format(bounds[0]),
            bin_right="{:.0f}".format(bounds[1]),
        )
        if i in (0, len(bins) - 2):
            label = end_label_fmt.format(
                word=word, bin_left="{:.0f}".format(min(np.abs(bounds)))
            )
        bin_class = f"{bounds[0] * multiplier:+.1f}_to_{bounds[1]  * multiplier:+.1f}"
        bin_classes.append(bin_class)
        bin_labels.append(label)
        bin_centres.append(np.mean(np.asarray(bounds) * multiplier))

    label_func = lambda x: pd.cut([x], bins, labels=bin_labels)[0]
    class_func = lambda x: pd.cut([x], bins, labels=bin_classes)[0]

    return {
        "bins": bins,
        "bins_left": bins[:-1],
        "bins_right": bins[1:],
        "bin_centres": bin_centres,
        "labels": bin_labels,
        "classes": bin_classes,
        "min": range_min,
        "max": range_max,
        "extrema": extrema,
        "step": step,
        "label_func": label_func,
        "class_func": class_func,
    }


def calculate_salinity_indicator_results(
    tdsann,
    dfn,
    pct_diff=None,
    trend_pct=None,
    years="final",
    regression_method="least-absolute-deviation",
    dropna=True,
    **kwargs,
):
    """Calculate 'pct_diff'/'curr_tds_pct_diff_indicator' and 'trend_pct'/'tds_trend_pct_change_indicator' values
    from annual salinity data.

    Args:
        tdsann (pandas.DataFrame): table of annual salinity data. It should have columns "tds" (floats),
            "collected_year" (int) and "well_id" (string or number) and will be passed as the first
            argument to :func:`wrap_technote.get_annual_mean_salinity_data`
        dfn (pandas.Series): recordO

    Returns: dict with keys "df" (pd.DataFrame), "data" (dict), "pct_diff" (dict) and
        "trend_pct" (dict)

    """
    data = {}
    for well_id in tdsann.well_id.unique():
        logger.debug(f"Calculating salinity indicators for {well_id}")
        df = get_annual_mean_salinity_data(tdsann, well_id)
        years, dfs, results = calculate_historical_salinity_trends(
            df, years="final", regression_method="least-absolute-deviation", **kwargs
        )
        if len(results):
            data[well_id] = {
                "years": years,
                "dfs": dfs,
                "results": results,
            }

    if pct_diff is None:
        pct_diff = generate_salinity_bins(
            range_min=float(dfn.pct_diff_range_min),
            range_max=float(dfn.pct_diff_range_max),
            step=float(dfn.pct_diff_step),
            word_positive="above",
            word_negative="below",
            label_fmt="{bin_left}% to {bin_right}% {word} mean",
            end_label_fmt="More than {bin_left}% {word} mean",
        )
    if trend_pct is None:
        trend_pct = generate_salinity_bins(
            range_min=float(dfn.trend_pct_range_min),
            range_max=float(dfn.trend_pct_range_max),
            step=float(dfn.trend_pct_step),
        )

    all_results = []

    for well_id, welldata in data.items():
        r = welldata["results"].iloc[-1]
        well_result = {
            "well_id": well_id,
            "mean_tds": welldata["dfs"][-1]["df"].iloc[-1]["mean"],
            "curr_tds": welldata["dfs"][-1]["df"].iloc[-1].tds,
            "curr_tds_pct_diff": r.current_pct_diff,
            "curr_tds_pct_diff_indicator": pct_diff["label_func"](r.current_pct_diff),
            "slope_pct_change_yr": r.slope_pct_change_yr,
            "slope_pct_change_trend_pd": r.slope_pct_change_yr * dfn.trend_length_years,
            "tds_trend_pct_change_indicator": trend_pct["label_func"](
                r.slope_pct_change_yr * dfn.trend_length_years
            ),
        }
        all_results.append(well_result)

    alldf = pd.DataFrame(all_results)
    if len(alldf) == 0:
        alldf = pd.DataFrame(
            columns=[
                "well_id",
                "mean_tds",
                "curr_tds",
                "curr_tds_pct_diff",
                "curr_tds_pct_diff_indicator",
                "slope_pct_change_yr",
                "slope_pct_change_trend_pd",
                "tds_trend_pct_change_indicator",
                "include_curr_pct_diff",
                "include_trend_pct",
                "validation_comment",
                "last_run_by",
                "last_run_date",
            ]
        )
    if dropna:
        alldf = alldf.dropna(
            subset=["curr_tds_pct_diff_indicator", "tds_trend_pct_change_indicator"],
            how="all",
        )
    alldf["include_curr_pct_diff"] = True
    alldf["include_trend_pct"] = True

    (
        curr_pct_diffs,
        tds_trend_pct_changes,
    ) = calculate_salinity_indicator_summary_results(alldf, pct_diff, trend_pct)

    return {
        "df": alldf,
        "data": data,
        "pct_diff": pct_diff,
        "trend_pct": trend_pct,
        "curr_pct_diffs": curr_pct_diffs,
        "tds_trend_pct_changes": tds_trend_pct_changes,
    }


def calculate_salinity_indicator_summary_results(alldf, pct_diff, trend_pct):
    """Calculate summary tables for *curr_pct_diffs* and *tds_trend_pct_changes*.

    Args:
        alldf (pd.DataFrame):
        pct_diff (bins)
        trend_pct (bins):

    Returns: tuple *curr_pct_diffs* and *tds_trend_pct_changes* (summary tables).

    """
    incl_idx = alldf.include_curr_pct_diff == True
    curr_pct_diffs = alldf[incl_idx].curr_tds_pct_diff_indicator.value_counts()
    curr_pct_diffs = curr_pct_diffs.reindex(pct_diff["labels"])
    curr_pct_diffs = curr_pct_diffs[::-1]
    curr_pct_diffs.index.name = "curr_tds_pct_diff_indicator"
    curr_pct_diffs.name = "n_wells"
    curr_pct_diffs = curr_pct_diffs.reset_index()

    pct_wells = curr_pct_diffs.n_wells / curr_pct_diffs.n_wells.sum() * 100
    pct_wells[np.isnan(pct_wells)] = 0

    logger.debug(f"pct_wells = {pct_wells}")
    curr_pct_diffs["pct_wells"] = round_to_100_percent(pct_wells)

    incl_idx2 = alldf.include_trend_pct == True
    tds_trend_pct_changes = alldf[
        incl_idx2
    ].tds_trend_pct_change_indicator.value_counts()
    tds_trend_pct_changes = tds_trend_pct_changes.reindex(trend_pct["labels"])
    tds_trend_pct_changes = tds_trend_pct_changes[::-1]
    tds_trend_pct_changes.index.name = "tds_trend_pct_change_indicator"
    tds_trend_pct_changes.name = "n_wells"
    tds_trend_pct_changes = tds_trend_pct_changes.reset_index()

    pct_wells = (
        tds_trend_pct_changes.n_wells / tds_trend_pct_changes.n_wells.sum() * 100
    )
    tds_trend_pct_changes["pct_wells"] = round_to_100_percent(pct_wells)

    curr_pct_diffs = curr_pct_diffs.fillna(0)
    tds_trend_pct_changes = tds_trend_pct_changes.fillna(0)

    return curr_pct_diffs, tds_trend_pct_changes


def calculate_historical_pct_diff_values(tdsann, pct_diff):
    """Calculate historical salinity percent difference from mean values.

    Args:
        tdsann (pd.DataFrame): annual salinities - must have "well_id", "collected_year"
            and "tds" columns
        pct_diff (dict): definition of percentage salinity difference bins

    Results: modifies *tdsann* in-place with new columns

    """
    for well_id in tdsann.well_id.unique():
        idx = tdsann.well_id == well_id
        mean_tds = tdsann[idx].tds.mean()
        tdsann.loc[idx, "pct_diff"] = ((tdsann[idx].tds - mean_tds) / mean_tds) * 100
        tdsann.loc[idx, "pct_diff_label"] = [
            "above" if v > 0 else "below" for v in tdsann[idx].pct_diff
        ]
        tdsann.loc[idx, "tds_indicator"] = [
            pct_diff["label_func"](v) for v in tdsann[idx].pct_diff
        ]

    return tdsann


def collate_salinity_summary_data(resource, db=None):
    """Collate salinity data outputs for a resource.

    Args:
        resource (Resource object)
        db (SAGeodataConnection object): optional, will
            be created if omitted.

    Returns:
        dict: details below

    The keys of the returned dictionary are:

    - current_tds_year
    - current_wells
    - current_wells_ex_trends
    - trend_wells
    - excluded_wells
    - curr_pct_diff_wells
    - tds_trend_pct_wells
    - resource
    - wells
    - wells_html
    - data_val
    - tds_changes
    - anntds_summ
    - anntds
    - tdstrends_mgl_summ_1
    - tdstrends_mgl_summ_2
    - tdstrends
    - tdstrends_5yr_summ_1
    - tdstrends_5yr_summ_2
    - curr_pct_diff
    - curr_pct_diff_summ
    - curr_pct_diff_summ_1
    - trend_pct
    - trend_pct_summ
    - trend_pct_summ_1
    - indicators_all

    """
    if db is None:
        db = gd.sageodata()

    current_tds_year = resource.trend_dfn.end_year

    s4 = resource.read_data("validated_data", sheet_name=None)
    s5 = resource.read_data("salinity_trends", sheet_name=None)
    s6 = resource.read_data("salinity_long_term_change", sheet_name=None)
    s7 = resource.read_data("salinity_indicators", sheet_name=None)

    well_ids = sorted(
        set(
            list(s4["valid data"].well_id.unique())
            + list(s4["invalid data"].well_id.unique())
        )
    )

    anntds = s4["current_mean_tds"]
    tdstrends = s5["final_trends"]
    current_wells = list(s4["valid data"].well_id.unique())
    trend_wells = list(s5["final trends"].well_id.unique())
    current_wells_ex_trends = set(current_wells) - set(trend_wells)
    excluded_wells = set(well_ids) - set(current_wells + trend_wells)
    curr_pct_diff_wells = list(s7["curr_pct_diff"].well_id.unique())
    tds_trend_pct_wells = list(s7["trend_pct"].well_id.unique())

    wells_found = db.find_wells(str(well_ids))
    if len(wells_found) > 0:
        wells = db.drillhole_details(wells_found).sort_values("well_id")
        wells["charts_link"] = wells.apply(
            lambda x: f"<a href='#charts-{x.well_id}'>{x.well_id}</a>", axis="columns"
        )
    else:
        wells = pd.DataFrame(columns=["well_id"])

    templates = tn.load_html_templates()

    data_val = resource.filter_table_by_well_ids(
        wells.well_id.values, table="Data_validation"
    )
    data_val_cols = [
        "well_id",
        "database",
        "start_period",
        "end_period",
        "action",
        "comment",
        "username",
    ]
    if len(data_val) > 0:
        data_val = data_val[data_val_cols]
    else:
        data_val = pd.DataFrame(columns=data_val_cols)

    anntds_summ = anntds[["tds"]].describe().round(2).T
    anntds = anntds.sort_values("tds")[["well_id", "unit_hyphen", "dh_name", "tds"]]

    remap_cols = {"sal_change": "5yr_change_mgL", "sal_pct_change": "5yr_change_in_pct"}
    cols_a = ["count", "mean", "std", "min", "25%", "50%", "75%", "max"]
    null_row_a = {col: pd.NA for col in cols_a}
    cols_b = [
        ("5yr_change_mgL", "count"),
        ("5yr_change_mgL", "mean"),
        ("5yr_change_mgL", "std"),
        ("5yr_change_mgL", "min"),
        ("5yr_change_mgL", "25%"),
        ("5yr_change_mgL", "50%"),
        ("5yr_change_mgL", "75%"),
        ("5yr_change_mgL", "max"),
        ("5yr_change_in_pct", "count"),
        ("5yr_change_in_pct", "mean"),
        ("5yr_change_in_pct", "std"),
        ("5yr_change_in_pct", "min"),
        ("5yr_change_in_pct", "25%"),
        ("5yr_change_in_pct", "50%"),
        ("5yr_change_in_pct", "75%"),
        ("5yr_change_in_pct", "max"),
    ]
    if len(tdstrends):
        tdstrends_mgl_summ_1 = (
            tdstrends.groupby(["status_change"])[["slope_yr"]].describe().round(2)
        )
        tdstrends_mgl_summ_2 = tdstrends[["slope_yr"]].describe().round(2).T
        tdstrends_5yr_summ_1 = (
            tdstrends.rename(columns=remap_cols)
            .groupby(["status_change"])[["5yr_change_mgL", "5yr_change_in_pct"]]
            .describe()
            .round(2)
        )
        tdstrends_5yr_summ_2 = (
            tdstrends[["sal_change", "sal_pct_change"]]
            .rename(columns=remap_cols)
            .describe()
            .round(2)
            .T
        )
        tdstrends = tdstrends.sort_values(["status_change", "sal_pct_change"])
        tdstrends = tdstrends[
            [
                "well_id",
                "status_change",
                "slope_yr",
                "sal_change",
                "sal_pct_change",
                "n_trend_year_obs",
            ]
        ].rename(columns=remap_cols)
    else:
        tdstrends_mgl_summ_1 = pd.DataFrame(columns=cols_a)
        tdstrends_mgl_summ_2 = pd.DataFrame(columns=cols_a)
        tdstrends_5yr_summ_1 = pd.DataFrame(columns=cols_b)
        tdstrends_5yr_summ_2 = pd.DataFrame(columns=cols_a)
    tdschange = s6["TDS changes"]

    curr_pct_diff_summ = s7["curr_pct_diff_summary"]
    trend_pct_summ = s7["tds_trend_pct_changes_summary"]
    curr_pct_diff = s7["curr_pct_diff"]
    if len(curr_pct_diff):
        curr_pct_diff_summ_1 = (
            curr_pct_diff["curr_tds_pct_diff"].describe().to_frame().T
        )
    else:
        curr_pct_diff_summ_1 = pd.DataFrame([null_row_a], index=["curr_tds_pct_diff"])

    trend_pct = s7["trend_pct"]
    if len(trend_pct):
        trend_pct_summ_1 = (
            trend_pct[["slope_pct_change_yr", "slope_pct_change_trend_pd"]].describe().T
        )
    else:
        trend_pct_summ_1 = pd.DataFrame(
            [null_row_a, null_row_a],
            index=["slope_pct_change_yr", "slope_pct_change_trend_pd"],
        )
    indicators_all = s7["all_results"]
    indicators_all[
        "validation_comment"
    ] = indicators_all.validation_comment.str.replace("\n", "<br />")

    return {
        "current_tds_year": current_tds_year,
        "current_wells": sorted(current_wells),
        "current_wells_ex_trends": sorted(current_wells_ex_trends),
        "trend_wells": sorted(trend_wells),
        "excluded_wells": sorted(excluded_wells),
        "curr_pct_diff_wells": sorted(curr_pct_diff_wells),
        "tds_trend_pct_wells": sorted(tds_trend_pct_wells),
        "resource": resource,
        "wells": wells.to_dict(orient="records"),
        "wells_html": wells.to_html(escape=False),
        "data_val": data_val,
        "tds_changes": s6["TDS changes"].to_html(escape=False),
        "anntds_summ": anntds_summ,
        "anntds": anntds,
        "tdstrends_mgl_summ_1": tdstrends_mgl_summ_1,
        "tdstrends_mgl_summ_2": tdstrends_mgl_summ_2,
        "tdstrends": tdstrends,
        "tdstrends_5yr_summ_1": tdstrends_5yr_summ_1,
        "tdstrends_5yr_summ_2": tdstrends_5yr_summ_2,
        "curr_pct_diff_summ": curr_pct_diff_summ,
        "curr_pct_diff": curr_pct_diff,
        "curr_pct_diff_summ_1": curr_pct_diff_summ_1,
        "trend_pct_summ": trend_pct_summ,
        "trend_pct": trend_pct,
        "trend_pct_summ_1": trend_pct_summ_1,
        "indicators_all": indicators_all,
    }


def construct_salinity_template_sentences(
    resource, data, highlight_method="papayawhip"
):
    """Construct template sentences for water level sections.

    Args:
        resource (wrap_technote.Resource): resource to summaries
        data (dict): collated salinity results - see
            :func:`wrap_technote.collate_salinity_summary_data`
        highlight_method (str): "papayawhip" by default -
            see :func:`wrap_technote.utils.highlight_fields` for
            details

    Returns:
        list: list of sentences

    """
    tds = data["anntds_summ"].iloc[0]
    cat1 = data["trend_pct_summ"].fillna(0)
    cat = cat1.groupby(cat1.tds_trend_pct_change_indicator.str.contains("increase"))[
        ["n_wells", "pct_wells"]
    ].sum()
    cat.index.name = "category"
    cat = cat.reset_index()
    cat.loc[cat.category == False, "category"] = "decreasing"
    cat.loc[cat.category == True, "category"] = "increasing"
    maj = cat.sort_values("n_wells", ascending=False).iloc[0]

    a_vs_an = lambda w: "an" if w[0].lower() in "aeiouh" else "a"
    tr_dirn = lambda v: "an increase" if v > 0 else "a decrease"

    trend_length_years_word = decimal_words[
        resource.dfn_salinity_indicators.trend_length_years
    ]
    trend_length_years_word_upper = (
        trend_length_years_word[0].upper() + trend_length_years_word[1:]
    )

    sentence_1 = (
        f"In <|{data['current_tds_year']}|>, sampling results from "
        f"<|{tds['count']:.0f}|> wells in the "
        f"<|{resource.report_resources_mapping.tech_note_sentence_value}|> "
        f"ranged between "
        f"<|{tds['min']:.0f}|> mg/L and <|{tds['max']:.0f}|> mg/L "
        f"with a median of <|{tds['50%']:.0f}|> mg/L (Figure XYZ)."
    )

    sentence_2 = (
        f"In the <|{trend_length_years_word}|> years to "
        f"<|{resource.dfn_salinity_indicators.trend_end_year}|>, "
        f"<|{maj['n_wells']:.0f}|> of <|{cat.n_wells.sum():.0f}|> wells "
        f"(<|{maj['pct_wells']:.0f}|>%) show <|{a_vs_an(maj.category)}|> "
        f"<|{maj.category}|> trend in salinity (Section XYZ; Figure XYZ)."
    )

    cat2 = data["trend_pct_summ_1"].loc["slope_pct_change_yr"].fillna(0)

    sentence_3 = (
        f"<|{trend_length_years_word_upper}|>-year trends show that "
        f"rates of change in salinity vary from a "
        f"<|{tr_dirn(cat2['min'])}|> of <|{abs(cat2['min']):.1f}|>% per year to "
        f"<|{tr_dirn(cat2['max'])}|> of <|{abs(cat2['max']):.1f}|>% per year, with a median rate of "
        f"<|{abs(cat2['50%']):.1f}|>% <|{tr_dirn(cat2['50%']).split()[1]}|> per year."
    )

    sentences = [sentence_1, sentence_2, sentence_3]
    sentences = tn.highlight_fields(sentences, highlight_method=highlight_method)
    return sentences



import sys
from datetime import datetime
import logging
import os
import textwrap
import shutil

import click
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import geopandas as gpd

import dew_gwdata as gd  # Access to SA Geodata database
import wrap_technote as tn  # GW data analysis functions


from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


logger = tn.get_logger()


@click.command()
@click.option("-p", "--reporting-period", default=tn.CURRENT_RPERIOD)
@click.option("-v", "--verbose", count=True)
@click.option("-s4", "--step4", is_flag=True)
@click.option("-s5", "--step5", is_flag=True)
@click.option("-s6", "--step6", is_flag=True)
@click.option("-s7", "--step7", is_flag=True)
@click.option("-s", "--static", is_flag=True)
@click.option("-r", "--report", default="")
@click.option("--figures/--no-figures", is_flag=True, default=True)
@click.argument("resource", required=False)
def salinities(
    resource,
    reporting_period,
    verbose,
    step4,
    step5,
    step6,
    step7,
    static,
    report,
    figures,
):
    logger.warning(f"wraptn.__version__ == {tn.__version__}")
    return salinities_run(
        resource,
        reporting_period,
        verbose,
        step4,
        step5,
        step6,
        step7,
        static,
        report,
        figures,
    )


def salinities_run(
    resource,
    reporting_period,
    verbose=0,
    step4=True,
    step5=True,
    step6=True,
    step7=True,
    static=True,
    report="",
    figures=True,
):
    handlers = []
    if verbose == 1:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
    if verbose == 2:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
        handlers.append({"sink": "wraptn_salinities.log", "level": "DEBUG"})
    config = {
        "handlers": handlers,
    }
    logger.configure(**config)

    logger.info(f"Using dew_gwdata {gd.__version__}")
    logger.info(f"Using wrap_technote {tn.__version__}")

    plt.rcParams["figure.dpi"] = 105

    db = gd.sageodata()

    if not resource:
        if report:
            logger.info(f"Finding resources for report: {report}")
            report = tn.Report(report, reporting_period)
            df = report.read_table("Report_Resources_mapping")
            resource_keys = [
                r for r in df[df.param == "TDS"].resource_key.unique() if r
            ]
            logger.info(f"Found resources: {resource_keys}")
    else:
        resource_keys = [resource]

    for resource in resource_keys:
        logger.info(
            f"Loading resource {resource} for reporting_period {reporting_period}"
        )
        resource = tn.Resource(resource, reporting_period)

        r = {}
        if step4:
            step4_data_validation(resource, db, plot_figures=figures)
        if step5:
            step5_salinity_trends(resource, db, plot_figures=figures)
        if step6:
            step6_salinity_change(resource, db, plot_figures=figures)
        if step7:
            step7_salinity_indicators(resource, db, plot_figures=figures)

        if static:
            create_static_salinity_page(resource, db)


def step4_data_validation(resource, db, plot_figures=True):
    current_tds_year = resource.trend_dfn.end_year

    resource_key = resource.resource_key
    logger.info(f"Finding wells for {resource_key}...")
    wells = resource.find_wells()
    logger.debug(f"Found wells: {wells}")

    logger.info("Retrieving salinity data...")
    sals0 = db.salinities(wells)
    sals0["database"] = "SA Geodata"

    logger.info("Remove all wells with < 3 observations...")
    sals0 = sals0.groupby("well_id").filter(lambda x: len(x) >= 3)

    logger.info("Filter out or refresh invalidated data...")
    sals, removals = tn.filter_tds_observations(
        sals0, return_removals=True, qc=resource
    )

    logger.info(f"removing any wells with no current data (in {current_tds_year}")
    keep = []
    removing = [removals]
    for well_id, well_sals in sals.groupby("well_id"):
        if current_tds_year in well_sals.collected_date.dt.year.unique():
            keep.append(well_sals)
        else:
            remove = pd.DataFrame(well_sals)
            remove["reason"] = f"No data from current year ({current_tds_year})"
            removing.append(remove)
    removals = pd.concat(removing)
    sals = pd.concat(keep)

    if plot_figures:
        for fn in resource.static_path.glob("plot_tds_data_validation*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)
        logger.info("Plot salinity data validation charts...")
        axes = tn.plot_tds_data_validation(
            sals,
            removals,
            show_comments=True,
            adjust_comments=False,
            savefig_and_close=True,
            path=resource.static_path,
        )

    logger.info(
        f"Selecting valid current year salinities with collected_date.dt.year == {current_tds_year}"
    )
    cols = [
        "well_id",
        "dh_no",
        "unit_long",
        "unit_hyphen",
        "obs_no",
        "dh_name",
        "easting",
        "northing",
        "zone",
        "latitude",
        "longitude",
        "aquifer",
    ]
    curr_sal_multiples = sals[sals.collected_date.dt.year == current_tds_year]
    curr_sal = (
        curr_sal_multiples.groupby(["well_id"], as_index=False).tds.mean().reset_index()
    )
    curr_sal = pd.merge(
        curr_sal_multiples[cols].drop_duplicates(),
        curr_sal,
        on="well_id",
        how="inner",
    )

    logger.debug(f"curr_sal ('current_mean_tds') {len(curr_sal)} records")

    output_data = {
        "valid_data": sals,
        "invalid_data": removals,
        "current_mean_tds": curr_sal,
    }
    resource.write_data_sheets("validated_data", output_data)

    for path in [resource.data_path]:
        spath = path / "spatial_outputs"
        spath.mkdir(exist_ok=True)
        tn.df_to_shp(curr_sal, spath / f"current_mean_tds_{resource.resource_key}.shp")

    return output_data


def step5_salinity_trends(resource, db, plot_figures=True):
    resource_key = resource.resource_key

    all_data = resource.read_data("validated_data", sheet_name=None)
    tdsdata = all_data["valid data"]
    invalid_tdsdata = all_data["invalid data"]
    # tdsdata = resource.read_data("validated_data", sheet_name="valid data")
    wells = db.drillhole_details(tdsdata.dh_no.unique())

    logger.info(
        "Add decimal number of days to the salinity observation table, to make linear regression easier."
    )

    if "collected_date" in tdsdata:
        tdsdata["ndays"] = tdsdata.collected_date.apply(tn.date_to_decimal)

    logger.info("Remove any wells that don't have data from the trend period.")

    tdsdata = resource.filter_tds_to_wells_with_trend_period_data(tdsdata)

    logger.info("Create a data frame that only has data from the trend period in it")

    trenddf = (
        tdsdata.groupby("well_id", as_index=False)
        .apply(lambda wdf: resource.filter_tds_to_trend_period(wdf))
        .reset_index(drop=True)
    )
    if len(trenddf) == 0:
        cols = list(tdsdata.columns.values)
        cols += [
            "slope_yr",
            "sal_change",
            "sal_pct_change",
            "status_change",
            "status",
            "status_combined",
        ]
        trenddf = pd.DataFrame(
            columns=cols,
        )

    logger.info("Calculate annual mean salinity within the trend period.")

    if len(trenddf):
        ann_mean_df = (
            trenddf.groupby(
                ["well_id", trenddf.collected_date.dt.year], as_index=False
            )[["ndays", "collected_date", "tds"]]
            .mean()
            .reset_index()
            .rename(columns={"collected_date": "collected_year"})
            .set_index("well_id")
        )
    else:
        ann_mean_df = pd.DataFrame(
            columns=list(tdsdata.columns.values)
            + ["collected_year", "ndays", "tds_max_date", "tds_max"]
        ).set_index("well_id")

    if len(ann_mean_df):
        ann_mean_df["collected_date"] = ann_mean_df.ndays.apply(tn.decimal_to_date)
        extrema_map = (
            trenddf[["well_id", "tds_max", "tds_max_date"]]
            .set_index("well_id")
            .drop_duplicates()
        )
        ann_mean_df["tds_max_date"] = ann_mean_df.index.map(extrema_map.tds_max_date)
        ann_mean_df["tds_max"] = ann_mean_df.index.map(extrema_map.tds_max)
    else:
        extrema_map = pd.DataFrame(
            columns=["well_id", "tds_max", "tds_max_date"]
        ).set_index("well_id")

    ann_mean_df = ann_mean_df.reset_index()

    logger.info(
        "Flat out remove all wells with only one data point, because we can't calculate a trend at all."
    )

    num_data = ann_mean_df.groupby("well_id").collected_year.count()

    wells_n1 = num_data[num_data == 1].index.values
    tdsdata = tdsdata[lambda x: ~x.well_id.isin(wells_n1)]
    ann_mean_df = ann_mean_df[~ann_mean_df.well_id.isin(wells_n1)]
    well_ids = ann_mean_df.well_id.unique()

    logger.info(
        "Calculate how many observations fall within the trend period and exclude any if needed."
    )

    vals = resource.read_table("Data_validation")
    excl = vals.query("action == 'Exclude well trend'")
    incl = vals.query("action == 'Include well trend'")
    trendqc = (
        ann_mean_df.groupby("well_id").tds.count().to_frame(name="n_trend_year_obs")
    )
    trendqc["include"] = True
    trendqc["exclusion_reason"] = ""

    for well_id in num_data[num_data < resource.trend_dfn.min_data_pts].index:
        trendqc.loc[well_id, "include"] = False
        trendqc.loc[
            well_id, "exclusion_reason"
        ] = f"Only {num_data.loc[well_id]} years with data points in trend period"

    for ix, row in excl.iterrows():
        for well_id in row.well_id.split(","):
            well_id = well_id.strip()
            if well_id in trendqc.index:
                trendqc.loc[well_id, "include"] = False
                trendqc.loc[
                    well_id, "exclusion_reason"
                ] = f"{row.username}: {row.comment}"
    for ix, row in incl.iterrows():
        for well_id in row.well_id.split(","):
            well_id = well_id.strip()
            if well_id in trendqc.index:
                trendqc.loc[well_id, "include"] = True
                x = trendqc.loc[well_id, "exclusion_reason"]
                if x:
                    x = f" (overriding {x}"
                trendqc.loc[well_id, "exclusion_reason"] = (
                    f"{row.username}: {row.comment}" + x
                )

    logger.info("Calculate the trend line by linear regression.")

    welltrends = ann_mean_df.groupby("well_id").apply(
        lambda wdf: resource.apply_trend(wdf)
    )
    if "well_id" in welltrends.columns:
        # This well_id comes from the Trend definition table and should be removed in
        # favour of the index, which contains the actual well ID.
        welltrends = (
            welltrends.drop("well_id", axis="columns")
            .reset_index()
            .set_index("well_id")
        )

    if plot_figures:
        for fn in resource.static_path.glob("plot_salinity_trend*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)
        logger.info("Plot all chart lines.")

        fig = plt.figure(figsize=(7, 3))
        for well_id in sorted(well_ids):
            valid = tdsdata.query(f'well_id == "{well_id}"')
            invalid = invalid_tdsdata.query(f'well_id == "{well_id}"')
            trend_sals = ann_mean_df.query(f'well_id == "{well_id}"')

            ax = fig.add_subplot(111)

            ax = tn.plot_salinity_trend(
                valid,
                trend_sals,
                trend_lines=welltrends,
                excluded_sals=invalid,
                override_year_span="all",
                ax=ax,
            )
            tt = ax._title_text.get_text()
            if trendqc.loc[well_id].include == False:
                exclusion_label = "\n".join(
                    textwrap.wrap(str(trendqc.loc[well_id].exclusion_reason), 50)
                )
                tt += f"\nExcluded: {exclusion_label}"
            else:
                tt += f" - included"
            ax._title_text.set_text(tt)
            fig.tight_layout()
            fn = str(resource.static_path / f"plot_salinity_trend_{well_id}.png")
            fig.savefig(fn, dpi=130)
            tn.trim_whitespace_from_image(fn, append=False)
            fig.clf()

    logger.info("Calculate trend results")

    trend_results = pd.merge(welltrends, trendqc, left_index=True, right_index=True)
    trend_results = trend_results[
        [x for x in trend_results.columns if not x.startswith("Unnamed")]
    ]

    cols = [
        "well_id",
        "dh_no",
        "unit_long",
        "unit_hyphen",
        "obs_no",
        "dh_name",
        "easting",
        "northing",
        "zone",
        "latitude",
        "longitude",
        "aquifer",
    ]
    trend_results = pd.merge(
        tdsdata[cols].drop_duplicates(),
        trend_results,
        on="well_id",
        how="inner",
    )

    if len(trend_results):
        trend_results.loc[
            trend_results["status_change"] == "Decreasing", "gsr_status_colour"
        ] = "Blue"
        trend_results.loc[
            trend_results["status_change"] == "Stable", "gsr_status_colour"
        ] = "Green"
        trend_results.loc[
            trend_results["status_change"] == "Increasing", "gsr_status_colour"
        ] = "Yellow"
        trend_results.loc[
            (trend_results["status_change"] == "Decreasing")
            & (trend_results["status_threshold"].str.startswith("Above")),
            "gsr_status_colour",
        ] = "Purple"
        trend_results.loc[
            (trend_results["status_change"] == "Stable")
            & (trend_results["status_threshold"].str.startswith("Above")),
            "gsr_status_colour",
        ] = "Grey"
        trend_results.loc[
            (trend_results["status_change"] == "Increasing")
            & (trend_results["status_threshold"].str.startswith("Above")),
            "gsr_status_colour",
        ] = "Black"

    logger.info("Write results to spreadsheet...")

    for col in ["status", "status_change", "status_combined", "include"]:
        if not col in trend_results:
            trend_results[col] = ""

    final_trends = trend_results[trend_results.include == True]
    excluded_trends = trend_results[trend_results.include == False]

    output_data = {
        "data_in_trend_period": trenddf,
        "mean_annual_tds_in_trend_pd": ann_mean_df,
        "final_trends": final_trends,
        "excluded_trends": excluded_trends,
        "data_quality": trendqc,
    }
    resource.write_data_sheets("salinity_trends", output_data)

    for path in [resource.data_path]:
        spath = path / "spatial_outputs"
        spath.mkdir(exist_ok=True)
        tn.df_to_shp(
            final_trends, spath / f"salinity_trends_{resource.resource_key}.shp"
        )

    return output_data


def step6_salinity_change(resource, db, plot_figures=True):
    data = resource.read_data("validated_data", sheet_name=None)
    params = resource.read_table("GW_TDS_long_term_changes")

    vdata = data["valid data"]
    if len(vdata):
        vdata["collected_year"] = vdata["collected_date"].dt.year

        wells = db.drillhole_details(data["valid data"].dh_no.unique())
        long_pd_end = pd.Timestamp(f"{int(params.period_end)}-12-31")
        long_pd_start = pd.Timestamp(f"{int(params.long_period_start)}-01-01")

        short_pd_end = pd.Timestamp(f"{int(params.period_end)}-12-31")
        short_pd_start = pd.Timestamp(f"{int(params.short_period_start)}-01-01")

        dt_col = "collected_date"
        long_pd = (vdata[dt_col] >= long_pd_start) & (vdata[dt_col] <= long_pd_end)
        short_pd = (vdata[dt_col] >= short_pd_start) & (vdata[dt_col] <= short_pd_end)

        short_vdata = vdata[short_pd]
        long_vdata = vdata[long_pd]

        short_nyears = short_vdata.groupby("well_id").collected_year.nunique()
        long_nyears = long_vdata.groupby("well_id").collected_year.nunique()

        obs_extent = (
            long_vdata.groupby("well_id")
            .collected_year.agg(["first", "last"])
            .rename(columns={"first": "first_year", "last": "last_year"})
        )

        nobs = pd.merge(
            long_nyears, short_nyears, left_index=True, right_index=True
        ).rename(
            columns={
                "collected_year_x": "long_pd_nyears_w_obs",
                "collected_year_y": "short_pd_nyears_w_obs",
            }
        )

        stats = pd.merge(
            obs_extent, nobs, left_index=True, right_index=True, how="inner"
        )
        stats = stats[stats.first_year < short_pd_start.year].sort_values(
            ["long_pd_nyears_w_obs", "short_pd_nyears_w_obs"], ascending=False
        )

        short_mean_tds = short_vdata.groupby("well_id").tds.mean()
        long_mean_tds = long_vdata.groupby("well_id").tds.mean()
        means = pd.merge(
            long_mean_tds,
            short_mean_tds,
            left_index=True,
            right_index=True,
            how="inner",
        ).rename(columns={"tds_y": "short_pd_tds", "tds_x": "long_pd_tds"})
        means2 = pd.merge(stats, means, left_index=True, right_index=True, how="inner")
        means2["tds_change"] = means2["short_pd_tds"] - means2["long_pd_tds"]
        means2["tds_pct_change"] = means2["tds_change"] / means2["short_pd_tds"] * 100
        means2 = means2.sort_values("tds_pct_change").round(
            {"long_pd_tds": 0, "short_pd_tds": 0, "tds_change": 1, "tds_pct_change": 2}
        )
        cols = [x for x in means2.columns]
        cols.insert(0, "short_period")
        cols.insert(0, "long_period")
        means2["short_period"] = f"{short_pd_start.year} to {short_pd_end.year}"
        means2["long_period"] = f"{long_pd_start.year} to {long_pd_end.year}"
        means2 = means2[cols].sort_values(
            ["long_pd_nyears_w_obs", "short_pd_nyears_w_obs"], ascending=False
        )
    else:
        means2 = pd.DataFrame(
            columns=[
                "well_id",
                "long_period",
                "short_period",
                "first_year",
                "last_year",
                "long_pd_nyears_w_obs",
                "short_pd_nyears_w_obs",
                "long_pd_tds",
                "short_pd_tds",
                "tds_change",
                "tds_pct_change",
            ]
        )
        short_vdata = vdata
        long_vdata = vdata

    means2 = means2.reset_index()

    output_data = {
        "tds_changes": means2,
        "short_period_tds_data": short_vdata,
        "long_period_tds_data": long_vdata,
    }
    resource.write_data_sheets("salinity_long_term_change", output_data)

    return output_data


def step7_salinity_indicators(resource, db, plot_figures=True):
    tds = resource.read_data("validated_data", "valid_data")

    logger.info(f"Reducing validated data to annual mean TDS")
    tdsann = tn.reduce_to_annual_tds(tds, reduction_func="mean")

    logger.info(f"Read Definitions_salinity_indicators")
    dfn = resource.read_table("Definitions_salinity_indicators").iloc[0]
    logger.debug(f"dfn =\n{dfn}")

    logger.info(f"Removing any TDS data > trend_end_year ({dfn.trend_end_year})")
    tdsann = tdsann[tdsann.collected_year <= int(dfn.trend_end_year)]

    logger.info(f"Calculating annual TDS statistics")
    annstats = tn.calculate_annual_tds_stats(tdsann)

    # Plot TDS resoruce monitoring history (mean/stdev salinity vs time for resource)
    logger.info(f"plot_tds_resource_monitoring_history for {resource.resource_key}")
    axes = tn.plot_tds_resource_monitoring_history(annstats)
    for path in [
        resource.data_path,
    ]:
        fn = str(
            path / f"{resource.resource_key}_plot_tds_resource_monitoring_history.png"
        )
        axes[0].figure.savefig(fn, dpi=130)

    # Plot TDS resource variability (mean vs stdev TDS with aspect of n_wells and year)
    logger.info(f"plot_tds_resource_variability for {resource.resource_key}")
    axes = tn.plot_tds_resource_variability(annstats)
    for path in [
        resource.data_path,
    ]:
        fn = str(path / f"{resource.resource_key}_plot_tds_resource_variability.png")
        axes[0].figure.savefig(fn, dpi=130)

    # Generate all salinity results
    logger.info(f"calculate_salinity_indicator_results for {resource.resource_key}")
    r = tn.calculate_salinity_indicator_results(
        tdsann, dfn, trend_length_years=dfn.trend_length_years
    )

    # Apply QC validations.
    r["df"], curr_pct_diff_qc = resource.get_salinity_curr_pct_diff_qc_results(
        r["df"], tdsdf=tdsann
    )
    r["df"], trend_pct_qc = resource.get_salinity_trend_pct_qc_results(
        r["df"], tdsdf=tdsann
    )
    r["df"] = resource.apply_salinity_indicator_validations(r["df"], conn=db)

    # Re-calculate summary tables.
    c, t = tn.calculate_salinity_indicator_summary_results(
        r["df"], r["pct_diff"], r["trend_pct"]
    )
    r["curr_pct_diffs"] = c
    r["tds_trend_pct_changes"] = t
    logger.info(f"Summary curr_pct_diff\n{r['curr_pct_diffs']}")
    logger.info(f"Summary tds_trend_pct_changes\n{r['tds_trend_pct_changes']}")

    # calculate historical pct_diff values
    logger.info(f"calculate_historical_pct_diff_values for {resource.resource_key}")
    tdsann = tn.calculate_historical_pct_diff_values(tdsann, r["pct_diff"])

    if plot_figures:
        for fn in resource.static_path.glob("salinity_status_and_trend_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)

        # charting individual well status and trend figures
        logger.info(
            f"plot_salinity_status_and_trend_simple_with_indicators for {resource.resource_key}"
        )
        for well_id, wdata in r["data"].items():
            logger.info(f"Plotting salinity status and trend for {well_id}")
            df = wdata["dfs"][-1]["df"]
            trend_df = wdata["dfs"][-1]["trend_df"]
            curr_result = wdata["results"].iloc[-1]
            ax = tn.plot_salinity_status_and_trend_simple_with_indicators(
                df,
                trend_df,
                curr_result,
                r["df"][lambda x: x.well_id == well_id].iloc[0],
                dfn,
            )
            fn = resource.static_path / f"salinity_status_and_trend_{well_id}.png"
            ax.figure.savefig(str(fn), dpi=130)

    if len(tdsann):
        # Plot historical TDS percent differences
        logger.info(f"plot_salinity_historical_pct_diffs for {resource.resource_key}")
        axes = tn.plot_salinity_historical_pct_diffs(
            tdsann,
            r["pct_diff"],
            included_wells=r["df"][
                r["df"].include_curr_pct_diff == True
            ].well_id.unique(),
        )
        for path in [
            resource.data_path,
        ]:
            fn = str(
                path / f"{resource.resource_key}_plot_salinity_historical_pct_diffs.png"
            )
            axes[0].grid(False)
            axes[0].figure.savefig(fn, dpi=130)

    if len(r["curr_pct_diffs"]):
        # plot_salinity_curr_pct_diffs_summary
        logger.info(f"plot_salinity_curr_pct_diffs_summary for {resource.resource_key}")
        ax = tn.plot_salinity_curr_pct_diffs_summary(
            r["curr_pct_diffs"], r["pct_diff"], style="internal", results=r["df"]
        )
        for path in [
            resource.data_path,
        ]:
            fn = str(
                path
                / f"{resource.resource_key}_plot_salinity_curr_pct_diffs_summary.png"
            )
            ax.figure.savefig(fn, dpi=130)

    if len(r["tds_trend_pct_changes"]):
        # plot_salinity_trend_pct_summary
        logger.info(f"plot_salinity_trend_pct_summary for {resource.resource_key}")
        ax = tn.plot_salinity_trend_pct_summary(
            r["tds_trend_pct_changes"],
            r["trend_pct"],
            style="internal",
            results=r["df"],
        )
        for path in [
            resource.data_path,
        ]:
            fn = str(
                path / f"{resource.resource_key}_plot_salinity_trend_pct_summary.png"
            )
            ax.figure.savefig(fn, dpi=130)

    output_data = {
        "annual_mean_tds": tdsann,
        "curr_pct_diff": r["df"][r["df"].include_curr_pct_diff == True],
        "trend_pct": r["df"][r["df"].include_trend_pct == True],
        "excluded_curr_pct_diff": r["df"][r["df"].include_curr_pct_diff == False],
        "excluded_trend_pct": r["df"][r["df"].include_trend_pct == False],
        "all_results": r["df"],
        "curr_pct_diff_summary": r["curr_pct_diffs"],
        "tds_trend_pct_changes_summary": r["tds_trend_pct_changes"],
        "qc_curr_pct_diff": curr_pct_diff_qc.reset_index(),
        "qc_trend_pct_qc": trend_pct_qc.reset_index(),
    }
    resource.write_data_sheets("salinity_indicators", output_data)

    return output_data


def create_static_salinity_page(resource, db):
    logger.info(f"Creating a static page for {resource.resource_key}")

    templates = tn.load_html_templates()

    rp = resource.reporting_period

    retdata = tn.collate_salinity_summary_data(resource, db)
    sentences = tn.construct_salinity_template_sentences(resource, retdata)
    retdata["sentences"] = sentences

    with open(str(resource.static_path / "salinities.html"), mode="w") as html:
        template = templates.get_template("static/salinities.html")
        html.write(template.render(retdata))



import sys
import logging
from pprint import pprint
import re
import os
import subprocess
import sys
from pathlib import Path

import click
import pandas as pd

logging.getLogger("matplotlib").setLevel(logging.CRITICAL)
logging.getLogger("fiona").setLevel(logging.CRITICAL)
logging.getLogger("PIL").setLevel(logging.CRITICAL)

import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib import colors as mcolors
from matplotlib import ticker as mticker
from matplotlib import patches as mpatches
import geopandas as gpd
import numpy as np
import pyproj

import dew_gwdata as gd
import ausweather
import wrap_technote as tn


logger = tn.get_logger()


@click.command()
@click.option("-p", "--reporting-period", default=tn.CURRENT_RPERIOD)
@click.option("-v", "--verbose", count=True)
@click.option("-k", "--resource-key", default=None)
@click.option("-r", "--report-key", default=None)
@click.option("--local/--no-local", default=True)
@click.option("--nbs/--no-nbs", default=True)
@click.option("--chart-glob", default=".*")
def summaries(
    reporting_period, report_key, verbose, resource_key, local, nbs, chart_glob
):
    return summaries_run(
        reporting_period, report_key, verbose, resource_key, local, nbs, chart_glob
    )


def summaries_run(
    reporting_period,
    report_key,
    verbose=0,
    resource_key=None,
    local=True,
    nbs=True,
    chart_glob=".*",
):
    handlers = []
    if verbose == 1:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
    if verbose == 2:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
        handlers.append({"sink": "wraptn_summaries.log", "level": "DEBUG"})
    config = {
        "handlers": handlers,
    }
    logger.configure(**config)

    logger.info(f"Using dew_gwdata {gd.__version__}")
    logger.info(f"Using wrap_technote {tn.__version__}")
    logger.info(f"Using ausweather {ausweather.__version__}")

    logger.info(f"Using reporting_period; {reporting_period}")

    logger.debug(f"chart_glob = {chart_glob}")

    plt.rcParams["figure.dpi"] = 150

    if resource_key is not None:
        if not resource_key == "none":
            report = tn.Report(report_key, reporting_period)
            logger.info(
                f"resource_key {resource_key} was specified; making summaries for that resource only."
            )
            make_summaries_for_resource(report, resource_key, chart_glob=chart_glob)

    if resource_key is None:
        logger.info(
            f"resource_key was not specified; making summaries for all resources in {report_key} report"
        )
        make_summaries_for_report(
            report_key, reporting_period, nbs=nbs, chart_glob=chart_glob
        )

    if resource_key is None or resource_key == "none":
        logger.info(f"Creating static HTML page for {report_key} report")
        create_static_summaries_page(report_key, reporting_period, local=local)


def create_static_summaries_page(report_key, reporting_period, local=False):
    logger.info(
        f"Creating static HTML page for {reporting_period}  {report_key} report"
    )
    report = tn.Report(report_key, reporting_period)
    templates = tn.load_html_templates()
    template = templates.get_template("static/gw_summary.html")

    resource_key_paths = {}
    resources = {}
    for section_name, section in report.report_settings.resources.items():
        logger.info(f"Working on section: {section_name}")
        for param, resource_keys in section.items():
            for resource_key in resource_keys:
                logger.info(
                    f"section={section_name} param={param} resource_key={resource_key}"
                )
                r = tn.Resource(resource_key, reporting_period)
                maptable = r.read_table("Report_Resources_mapping")
                logger.info(f"Adding link to {resource_key} static page...")
                if local:
                    path = maptable[maptable.param == param].path.iloc[0]
                    parts = Path(path).parts[2:]
                    p = Path("..")
                    for part in parts:
                        p = p / part
                else:
                    p = r.data_path
                if param == "WL":
                    resource_key_paths[resource_key] = (
                        p / "technote_static" / "waterlevels.html"
                    )
                elif param == "TDS":
                    resource_key_paths[resource_key] = (
                        p / "technote_static" / "salinities.html"
                    )
                resources[resource_key] = tn.load_resource(
                    resource_key, reporting_period
                )

    logger.info(f"Creating static page for {report_key}")
    if local:
        page_path = "."
    else:
        page_path = report.gw_summaries_path

    resource_sentences = {}
    for resource_key, resource in resources.items():
        if resource_key.endswith("WL"):
            retdata = tn.collate_waterlevel_summary_data(resource)
            sentences = tn.construct_waterlevel_template_sentences(resource, retdata)
            resource_sentences[resource_key] = sentences

        elif resource_key.endswith("TDS"):
            retdata = tn.collate_salinity_summary_data(resource)
            sentences = tn.construct_salinity_template_sentences(resource, retdata)
            resource_sentences[resource_key] = sentences

    with open(str(report.gw_summaries_path / "gw_summary.html"), mode="w") as html:
        html.write(
            template.render(
                {
                    "report": report,
                    "page_path": page_path,
                    "resource_key_paths": resource_key_paths,
                    "resource_sentences": resource_sentences,
                    "resources": resources,
                }
            )
        )


def make_summaries_for_report(report_key, reporting_period, nbs=True, chart_glob=".*"):
    logger.info(f"Making summaries for report: {report_key}")
    report = tn.Report(report_key, reporting_period)

    if nbs:
        path = report.gw_summaries_path.parent
        if path.glob("*.ipynb"):
            with tn.cd(path):
                subprocess.call(
                    [
                        "jupyter",
                        "nbconvert",
                        "--ExecutePreprocessor.kernel_name=python3",
                        "--execute",
                        "--to",
                        "html",
                        "*.ipynb",
                    ]
                )

    logger.info("Loading aggregate resources data.")
    wl_agg_ranks = report.get_aggregate_resources_data(
        "WL", "recovery_wl_data", "current ranked WLs"
    )
    wl_agg_trends = report.get_aggregate_resources_data(
        "WL", "recovery_wl_trends", "final trends"
    )
    tds_agg_trends = report.get_aggregate_resources_data(
        "TDS", "salinity_trends", "final trends"
    )

    for agg_key, df in wl_agg_ranks.items():
        for path in [report.gw_summaries_path]:
            fn = f"{agg_key} - recovery_wl_data__current_ranked_wls.csv"
            fn_path = path / fn
            df.to_csv(fn_path, index=False)

    for agg_key, df in wl_agg_trends.items():
        for path in [report.gw_summaries_path]:
            fn = f"{agg_key} - recovery_wl_trends__final_trends.csv"
            fn_path = path / fn
            df.to_csv(fn_path, index=False)

    for agg_key, df in tds_agg_trends.items():
        for path in [report.gw_summaries_path]:
            fn = f"{agg_key} - salinity_trends__final_trends.csv"
            fn_path = path / fn
            df.to_csv(fn_path, index=False)

    chart_glob_re = re.compile(chart_glob)

    if chart_glob_re.match("aggregate_current_WL_rankings"):
        logger.info("Plotting _aggregate_current_WL_rankings")
        filename = f"{report.report_key}_aggregate_current_WL_rankings.png"
        bcmap = tn.BoMClassesColormap()
        ax = tn.plot_internal_classes_for_aggregate_resources(
            wl_agg_ranks, bcmap.class_names, bcmap.colours_rgba(), col="rswl_bom_class"
        )
        ax.set_title(
            f"{report.report_key} Aggregate current WL rankings", fontsize="medium"
        )
        ax.figure.tight_layout()
        for path in [report.gw_summaries_path]:
            fn = str(path / filename)
            ax.figure.savefig(fn, dpi=150)
            tn.trim_whitespace_from_image(fn, append=False)

    if chart_glob_re.match("aggregate_5yr_WL_trends_triclass"):
        logger.info("Plotting _aggregate_5yr_WL_trends_triclass")
        filename = f"{report.report_key}_aggregate_5yr_WL_trends_triclass.png"
        bcmap = tn.BoMClassesColormap()
        ax = tn.plot_internal_classes_for_aggregate_resources(
            wl_agg_trends,
            tn.wl_status_changes,
            tn.wl_status_change_colours,
            col="status_change",
        )
        ax.set_title(f"{report.report_key} Aggregate 5-yr WL trends", fontsize="medium")
        ax.figure.tight_layout()
        for path in [report.gw_summaries_path]:
            fn = str(path / filename)
            ax.figure.savefig(fn, dpi=150)
            tn.trim_whitespace_from_image(fn, append=False)

    if chart_glob_re.match("aggregate_5yr_TDS_trends_triclass"):
        logger.info("Plotting _aggregate_5yr_TDS_trends_triclass")
        filename = f"{report.report_key}_aggregate_5yr_TDS_trends_triclass.png"
        ax = tn.plot_internal_classes_for_aggregate_resources(
            tds_agg_trends,
            tn.tds_status_changes,
            tn.tds_status_change_colours,
            col="status_change",
        )
        ax.set_title(
            f"{report.report_key} Aggregate 5-yr TDS trends", fontsize="medium"
        )
        ax.figure.tight_layout()
        for path in [report.gw_summaries_path]:
            fn = str(path / filename)
            ax.figure.savefig(fn, dpi=150)
            tn.trim_whitespace_from_image(fn, append=False)

    for section_name, section in report.report_settings.resources.items():
        logger.info(f"Working on report section: {section_name}")
        for param, resource_keys in section.items():
            logger.info(f"Iterating over {len(resource_keys)} {param} resource_keys:")
            for resource_key in resource_keys:
                logger.info(f"Loading settings for {resource_key}")
                try:
                    settings = report.resource_settings(resource_key)
                except ModuleNotFoundError:
                    logger.info("No settings found, moving to the next resource.")
                    continue
                else:
                    make_summaries_for_resource(
                        report, resource_key, chart_glob=chart_glob
                    )


def make_summaries_for_resource(report, resource_key, chart_glob=".*"):
    if resource_key.endswith("WL"):
        return make_wl_summaries(report, resource_key, chart_glob=chart_glob)
    elif resource_key.endswith("TDS"):
        return make_tds_summaries(report, resource_key, chart_glob=chart_glob)


def make_wl_summaries(report, resource_key, chart_glob=".*"):
    logger.info(
        f"Making WL summaries for report='{report.report_key}' resource='{resource_key}'"
    )
    resource = tn.Resource(
        resource_key, tn.paths_reporting_period[report.reporting_period.path]
    )
    s = report.resource_settings(resource_key)
    ranks_data = resource.read_data("recovery_wl_data", sheet_name=None)
    logger.info(f"recovery_wl_data sheet names: {ranks_data.keys()}")
    wlranks = ranks_data["ranked WLs"]
    all_wlranks = pd.concat([wlranks, ranks_data["ranks excl"]])
    ranks_qc = ranks_data["data quality"]
    logger.info(f"{len(wlranks.well_id.unique())} wells with historical ranked WLs")
    wlranks = pd.concat([wlranks, s.dry_rows])

    curr_ranks = ranks_data["current ranked WLs"]
    logger.info(f"{len(curr_ranks)} with current year WL rank")

    logger.debug(f"Compiling chart_glob = {chart_glob}")
    chart_glob_re = re.compile(chart_glob)

    if len(curr_ranks):
        if chart_glob_re.match("rankings"):
            ax = tn.plot_wl_ranking_classes(curr_ranks)
            ax.figure.tight_layout()
            filename = f"{resource_key}_rankings.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

            ax = tn.plot_wl_ranking_map(
                curr_ranks,
                s.ranking_map_elements,
                s.ranking_map_annotations,
                leg_frame=s.leg_frame,
                leg_loc=s.leg_loc,
                markersize=s.wl_ranking_marker_size,
            )
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            ax.set_xlim(
                x0 + s.ranking_map_xaxis_shift[0], x1 + s.ranking_map_xaxis_shift[1]
            )
            ax.set_ylim(
                y0 + s.ranking_map_yaxis_shift[0], y1 + s.ranking_map_yaxis_shift[1]
            )
            ax.set_xlim(*s.ranking_map_limits["x"])
            ax.set_ylim(*s.ranking_map_limits["y"])
            ax.figure.tight_layout()
            filename = f"{resource_key}_rankings_map.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

        if chart_glob_re.match("wl_rankings_internal"):
            logger.info("Charting: tn.plot_wl_rankings_internal")
            fig = tn.plot_wl_rankings_internal(curr_ranks, title=resource.resource_key)
            filename = f"{resource.resource_key}_wl_rankings_internal.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                fig.savefig(fn, dpi=120)
                tn.trim_whitespace_from_image(fn, append=False)

        if chart_glob_re.match("wl_historical_rankings"):
            logger.info("Charting:tn.plot_wl_historical_rankings")
            fig = tn.plot_wl_historical_rankings(
                all_wlranks, ranks_qc, title=resource.resource_key
            )
            filename = f"{resource.resource_key}_wl_historical_rankings.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                fig.savefig(fn, dpi=120)
                tn.trim_whitespace_from_image(fn, append=False)

    logger.info("Load trend data")
    wl_trend_data = resource.read_data("recovery_wl_trends", sheet_name=None)
    wl_trend_df = wl_trend_data["final trends"]
    wl_annual_threshold = resource.trend_dfn.ann_rate_threshold

    if len(wl_trend_df):
        if chart_glob_re.match("trend_triclass"):
            ax = tn.plot_wl_trend_triclass_bars(wl_trend_df)
            if s.trend_triclass_xticks:
                ax.set_xticks(s.trend_triclass_xticks)
            ax.figure.tight_layout()
            filename = f"{resource_key}_trend_triclass.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

            ax = tn.plot_wl_trend_triclass_map(
                wl_trend_df,
                s.trend_map_elements,
                s.trend_map_annotations,
                leg_frame=s.leg_frame,
                leg_loc=s.leg_loc,
                markersize=s.wl_trend_marker_size,
            )
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            ax.set_xlim(
                x0 + s.trend_map_xaxis_shift[0], x1 + s.trend_map_xaxis_shift[1]
            )
            ax.set_ylim(
                y0 + s.trend_map_yaxis_shift[0], y1 + s.trend_map_yaxis_shift[1]
            )
            ax.set_xlim(*s.trend_map_limits["x"])
            ax.set_ylim(*s.trend_map_limits["y"])
            ax.figure.tight_layout()

        if chart_glob_re.match("wl_trends_internal"):
            logger.info("Charting: tn.plot_wl_trends_internal")
            fig = tn.plot_wl_trends_internal(wl_trend_df, title=resource.resource_key)
            filename = f"{resource.resource_key}_wl_trends_internal.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                fig.savefig(fn, dpi=120)
                tn.trim_whitespace_from_image(fn, append=False)

        if chart_glob_re.match("trend_triclass_map"):
            filename = f"{resource_key}_trend_triclass_map.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)


def make_tds_summaries(report, resource_key, chart_glob=".*"):
    logger.info(
        f"Making TDS summaries for report='{report.report_key}' resource='{resource_key}'"
    )
    resource = tn.Resource(
        resource_key, tn.paths_reporting_period[report.reporting_period.path]
    )
    current_tds_year = resource.trend_dfn.end_year
    dfn = resource.read_table("Definitions_salinity_indicators").iloc[0]

    s = report.resource_settings(resource_key)
    valid_sal_sheet = resource.read_data("validated_data", sheet_name=None)
    valid_sal = valid_sal_sheet["valid data"]
    curr_sal = valid_sal_sheet["current_mean_tds"]

    chart_glob_re = re.compile(chart_glob)

    if len(curr_sal):
        if chart_glob_re.match("current"):
            tdsbins = tn.calculate_tds_bins(
                curr_sal,
                tds_bin_width=s.tds_bin_width,
                min_tds_colour=s.min_tds_colour,
                max_tds_colour=s.max_tds_colour,
                cmap=s.tds_cmap,
                min_tds_bin=s.min_tds_bin,
                max_tds_bin=s.max_tds_bin,
            )
            ax = tn.plot_tds_current_bars(
                tdsbins,
            )
            ax.figure.tight_layout()
            filename = f"{resource_key}_current.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

            ax = tn.plot_tds_current_map(
                tdsbins,
                curr_sal,
                s.current_map_elements,
                s.current_map_annotations,
                cmap=s.tds_cmap,
                leg_frame=s.leg_frame,
                leg_loc=s.leg_loc,
                markersize=s.tds_current_marker_size,
            )
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            ax.set_xlim(
                x0 + s.current_map_xaxis_shift[0], x1 + s.current_map_xaxis_shift[1]
            )
            ax.set_ylim(
                y0 + s.current_map_yaxis_shift[0], y1 + s.current_map_yaxis_shift[1]
            )
            ax.set_xlim(*s.current_map_limits["x"])
            ax.set_ylim(*s.current_map_limits["y"])
            ax.figure.tight_layout()
            filename = f"{resource_key}_current_map.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

        if chart_glob_re.match("tds_current_internal"):
            logger.info("Charting: tn.plot_tds_current_internal")
            ax = tn.plot_tds_current_internal(curr_sal, title=resource.resource_key)
            filename = f"{resource.resource_key}_tds_current_internal.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.tight_layout()
                ax.figure.savefig(fn, dpi=120)
                tn.trim_whitespace_from_image(fn, append=False)

    # Long-term salinity indicators
    sal_inds = resource.read_data("salinity_indicators", "all_results")
    sal_inds = pd.merge(
        sal_inds,
        valid_sal[["well_id", "longitude", "latitude"]].drop_duplicates(),
        on="well_id",
    )
    curr_pct_diff = sal_inds[sal_inds.include_curr_pct_diff == True]
    trend_pcts = sal_inds[sal_inds.include_trend_pct == True]
    curr_pct_diffs = resource.read_data("salinity_indicators", "curr_pct_diff_summary")
    trend_pct_changes = resource.read_data(
        "salinity_indicators", "tds_trend_pct_changes_summary"
    )

    if len(curr_pct_diff):
        if chart_glob_re.match("curr_pct_diff"):
            logger.info("Charting: tn.plot_tds_curr_pct_diff_bars")
            pct_diff = tn.generate_salinity_bins(
                range_min=float(dfn.pct_diff_range_min),
                range_max=float(dfn.pct_diff_range_max),
                step=float(dfn.pct_diff_step),
                word_positive="above",
                word_negative="below",
                label_fmt="{bin_left}% to {bin_right}% {word} mean",
                end_label_fmt="More than {bin_left}% {word} mean",
            )
            ax = tn.plot_tds_curr_pct_diff_bars(curr_pct_diffs, pct_diff)
            ax.figure.tight_layout()
            filename = f"{resource_key}_curr_pct_diff_bars.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

            logger.info("Charting: tn.plot_tds_curr_pct_diff_map")
            ax = tn.plot_tds_curr_pct_diff_map(
                curr_pct_diff,
                pct_diff,
                s.tds_curr_pct_diff_map_elements,
                s.tds_curr_pct_diff_map_annotations,
                leg_frame=s.leg_frame,
                leg_loc=s.leg_loc,
                markersize=s.tds_curr_pct_diff_marker_size,
            )
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            ax.set_xlim(
                x0 + s.tds_curr_pct_diff_map_xaxis_shift[0],
                x1 + s.tds_curr_pct_diff_map_xaxis_shift[1],
            )
            ax.set_ylim(
                y0 + s.tds_curr_pct_diff_map_yaxis_shift[0],
                y1 + s.tds_curr_pct_diff_map_yaxis_shift[1],
            )
            ax.set_xlim(*s.tds_curr_pct_diff_map_limits["x"])
            ax.set_ylim(*s.tds_curr_pct_diff_map_limits["y"])
            ax.figure.tight_layout()
            filename = f"{resource_key}_curr_pct_diff_map.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

    if len(trend_pcts):
        if chart_glob_re.match("trend_pct"):
            trend_pct = tn.generate_salinity_bins(
                range_min=float(dfn.trend_pct_range_min),
                range_max=float(dfn.trend_pct_range_max),
                step=float(dfn.trend_pct_step),
            )

            logger.info("Charting: tn.plot_tds_trend_pct_change_bars")
            ax = tn.plot_tds_trend_pct_change_bars(trend_pct_changes, trend_pct)
            ax.figure.tight_layout()
            filename = f"{resource_key}_tds_trend_pct_change_bars.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

            logger.info("Charting: tn.plot_tds_trend_pct_change_map")
            ax = tn.plot_tds_trend_pct_change_map(
                trend_pcts,
                trend_pct,
                s.tds_trend_pct_change_map_elements,
                s.tds_trend_pct_change_map_annotations,
                leg_frame=s.leg_frame,
                leg_loc=s.leg_loc,
                markersize=s.tds_trend_pct_change_marker_size,
            )
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            ax.set_xlim(
                x0 + s.tds_trend_pct_change_map_xaxis_shift[0],
                x1 + s.tds_trend_pct_change_map_xaxis_shift[1],
            )
            ax.set_ylim(
                y0 + s.tds_trend_pct_change_map_yaxis_shift[0],
                y1 + s.tds_trend_pct_change_map_yaxis_shift[1],
            )
            ax.set_xlim(*s.tds_trend_pct_change_map_limits["x"])
            ax.set_ylim(*s.tds_trend_pct_change_map_limits["y"])
            ax.figure.tight_layout()
            filename = f"{resource_key}_tds_trend_pct_change_map.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

    for path in [report.gw_summaries_path]:
        spath = path / "spatial_outputs"
        spath.mkdir(exist_ok=True)

        curr_pct_diff_gdf = gpd.GeoDataFrame(
            curr_pct_diff,
            geometry=gpd.points_from_xy(
                curr_pct_diff.longitude, curr_pct_diff.latitude
            ),
            crs="EPSG:7844",
        )
        if len(curr_pct_diff_gdf):
            tn.df_to_shp(
                curr_pct_diff_gdf,
                spath / f"salinity_curr_pct_diff_{resource.resource_key}.shp",
            )

        trend_pcts_gdf = gpd.GeoDataFrame(
            trend_pcts,
            geometry=gpd.points_from_xy(trend_pcts.longitude, trend_pcts.latitude),
            crs="EPSG:7844",
        )
        if len(trend_pcts_gdf):
            tn.df_to_shp(
                trend_pcts_gdf,
                spath / f"salinity_longterm_trend_pcts_{resource.resource_key}.shp",
            )

    sal_trend_sheet = resource.read_data("salinity_trends", sheet_name=None)
    sal_trends = sal_trend_sheet["final trends"]

    for path in [report.gw_summaries_path]:
        spath = path / "spatial_outputs"
        spath.mkdir(exist_ok=True)

        curr_gdf = gpd.GeoDataFrame(
            curr_sal,
            geometry=gpd.points_from_xy(curr_sal.longitude, curr_sal.latitude),
            crs="EPSG:7844",
        )
        if len(curr_gdf):
            tn.df_to_shp(
                curr_gdf, spath / f"salinity_current_mean_{resource.resource_key}.shp"
            )

        trend_gdf = gpd.GeoDataFrame(
            sal_trends,
            geometry=gpd.points_from_xy(sal_trends.longitude, sal_trends.latitude),
            crs="EPSG:7844",
        )
        if len(trend_gdf):
            tn.df_to_shp(
                trend_gdf,
                spath / f"salinity_triclass_trends_{resource.resource_key}.shp",
            )

    if len(sal_trends):
        if chart_glob_re.match("trend_triclass"):
            ax = tn.plot_tds_trend_triclass_bars(sal_trends)
            if s.trend_triclass_xticks:
                ax.set_xticks(s.trend_triclass_xticks)
            ax.figure.tight_layout()
            filename = f"{resource_key}_trend_triclass.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

            ax = tn.plot_tds_trend_triclass_map(
                sal_trends,
                s.trend_map_elements,
                s.trend_map_annotations,
                leg_frame=s.leg_frame,
                leg_loc=s.leg_loc,
                markersize=s.tds_trend_marker_size,
            )
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            ax.set_xlim(
                x0 + s.trend_map_xaxis_shift[0], x1 + s.trend_map_xaxis_shift[1]
            )
            ax.set_ylim(
                y0 + s.trend_map_yaxis_shift[0], y1 + s.trend_map_yaxis_shift[1]
            )
            ax.set_xlim(*s.trend_map_limits["x"])
            ax.set_ylim(*s.trend_map_limits["y"])
            ax.figure.tight_layout()
            filename = f"{resource_key}_trend_triclass_map.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                ax.figure.savefig(fn, dpi=250)
                tn.trim_whitespace_from_image(fn, append=False)

        if chart_glob_re.match("tds_trends_mgL_internal"):
            logger.info("Charting: tn.plot_tds_trends_internal mg/L")
            fig = tn.plot_tds_trends_internal(
                sal_trends, title=resource.resource_key, param="mg"
            )
            filename = f"{resource.resource_key}_tds_trends_mgL_internal.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                fig.savefig(fn, dpi=120)
                tn.trim_whitespace_from_image(fn, append=False)

        if chart_glob_re.match("tds_trends_5yr_pct_internal"):
            logger.info("Charting: tn.plot_tds_trends_internal 5yr_pct")
            fig = tn.plot_tds_trends_internal(
                sal_trends, title=resource.resource_key, param="5yr_pct"
            )
            filename = f"{resource.resource_key}_tds_trends_5yr_pct_internal.png"
            for path in [resource.data_path, report.gw_summaries_path]:
                fn = str(path / filename)
                fig.savefig(fn, dpi=120)
                tn.trim_whitespace_from_image(fn, append=False)



from datetime import datetime
import io
import logging
import os
import contextlib
from pathlib import Path
import sys

from loguru import logger as loguru_logger
import numpy as np
from PIL import Image, ImageChops
import geopandas as gpd
import pandas as pd


def get_logger():
    """Get a logging object (adds stderr for warning logger messages.

    Returns:
        loguru.Logger: logger suitable for use anywhere in wrap_technote.

    .. todo:: add sinks to a database for simple logging output.

    """
    from loguru import logger

    logger.remove()
    logger.add(sys.stderr, level="WARNING")
    return logger


logger = get_logger()


class Run:
    """Run an analysis.

    This is a parent class to be inherited. See, for example,
    :class:`wrap_technote.waterlevels.QueryAndValidateWaterLevels`.


    Args:
        log_to (func, or None): function which accept a single string.
        log_level (str): logging level to use.

    Other keyword arguments will be part of the initial data for the
    first step of the analysis; see subclasses.

    """

    def __init__(self, log_to=None, log_level="DEBUG", **kwargs):
        self.log_to = log_to
        self.log_level = log_level
        self.steps = []

        self.data = kwargs

    def if_then_use(self, data_variable_name, func):
        """Tool for constructing child classes.

        Args:
            data_variable_name (bool): key to self.data which should be tested
            func (function): the method to run or skip based on condition

        Returns:
            a function or a tuple of name, function - see :meth:`wrap_technote.Run.append_step`
            for more details on how this is parsed.

        """
        test_func = (
            lambda **kwargs: func(**kwargs) if self.data[data_variable_name] else {}
        )
        name = func.__name__
        return (name, test_func)

    def append_step(self, func):
        """Add an analysis step.

        Args:
            func (function or tuple of str, function): the analysis
                function; if it is an anonymous function (lambda), then
                you can provide a tuple where the first item is a string
                and the second item is the function.

        """
        if isinstance(func, tuple):
            name = func[0]
            function = func[1]
        else:
            function = func
            name = func.__name__
        step = Step(function, log_level=self.log_level, additional_log_to=self.log_to)
        step.name = name
        self.steps.append(step)

    def run_steps(self, first=0, last=None):
        """Execute analysis steps.

        Args:
            first (int): the first analysis step to execute (starting from
                zero)
            last (int): the last (inclusive) analysis step to execute

        Note that if you omit early step, it's likely later ones will
        fail due to the data not being generated and cached.

        """
        if last is None:
            last = len(self.steps) - 1

        logger.info(f"Running {self.__class__.__name__} steps {first} through {last}")
        for i, step in enumerate(self.steps):
            if i >= first and i <= last:
                logger.info(f"running step {i} {step.name}")
                step.run(self.data)
                logger.debug(
                    f"step {i} {step.name} returned data: {step.returned_data.keys()}"
                )
            self.data.update(step.returned_data)
        logger.info(f"Completed {self.__class__.__name__} steps {first} through {last}")


class Step:
    """Analysis step.

    Args:
        function
        log_level (str): logging to undertake.

    """

    def __init__(self, function, log_level="DEBUG", additional_log_to=None):
        self.log_level = log_level
        self.function = function
        self.name = self.function.__name__
        self.returned_data = {}
        self.log_messages = []
        self.additional_log_to = additional_log_to

    def log_message(self, msg):
        """Log a message both internally and to the additional logger if supplied."""
        self.log_messages.append(msg)
        if self.additional_log_to:
            self.additional_log_to(msg)

    def print_logs(self, output_stream=None, package=None):
        """Print log messages."""
        if output_stream is None:
            output_stream = sys.stdout
        messages = self.log_messages
        if package is None:
            messages = self.log_messages
        else:
            messages = [m for m in self.log_messages if package in m.splitlines()[0]]
        output_stream.write("\n".join(messages))

    def run(self, data):
        """Run the analysis step.

        This function adds a temporary logger to capture the step's details.

        """
        logger_id = loguru_logger.add(self.log_message, level=self.log_level)

        returned = self.function(**data)

        loguru_logger.remove(logger_id)

        if returned is None:
            returned = {}
        self.returned_data = returned


class DataFileDict(dict):
    """A dictionary where keys have spaces replaced by underscores
    transparently.

    """

    def __getitem__(self, key):
        key = key.lower().replace(" ", "_")
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        key = key.lower().replace(" ", "_")
        return super().__setitem__(key, value)


def intround(value):
    """Convert value to an integer."""
    return int(np.round(value))


def percentile_to_bom_class(values, fractional=False):
    """Convert percentile ranking to the BoM class.

    Args:
        values (sequence of numbers): the percentiles
        fractional (bool): whether the percentiles are 0-1 (True)
            or 0-100 (False).

    Returns:
        numpy.ndarray: string values either
        "Highest on record", "Very much above average", "Above average",
        "Average", "Below average", "Very much below average", or
        "Lowest on record"

    """
    values = np.asarray(values)
    if not fractional:
        values = values / 100.0
    min_value = np.nanmin(values)
    max_value = np.nanmax(values)
    results = []
    for value in values:
        if value == max_value:
            results.append("Highest on record")
        elif value == min_value:
            results.append("Lowest on record")
        else:
            if value < 0.1:
                results.append("Very much below average")
            elif value < 0.3:
                results.append("Below average")
            elif value < 0.7:
                results.append("Average")
            elif value < 0.9:
                results.append("Above average")
            elif value <= 1:
                results.append("Very much above average")
            elif np.isnan(value):
                results.append("No data")
            else:
                raise Exception(
                    "It should be impossible to get to this point in the code."
                )
    return np.asarray(results)


def map_percentile_into_bom_class(value, fractional=False):
    """Convert percentile to BoM classification.

    Args:
        value (float): the percentile - either between 0 and 100, or, if fractional
            is True, between 0 and 1
        fractional (bool): whether percentiles are out of 1 or 100

    Returns:
        str: the relevant BoM classification (out of 'Lowest on record',
        'Very much below average', 'Below average', 'Average',
        'Above average', 'Very much above average', or 'Highest on record').

    """
    if not fractional:
        value = value / 100.0
    if value == 0:
        return "Lowest on record"
    elif value < 0.1:
        return "Very much below average"
    elif value < 0.3:
        return "Below average"
    elif value < 0.7:
        return "Average"
    elif value < 0.9:
        return "Above average"
    elif value < 1:
        return "Very much above average"
    elif value == 1:
        return "Highest on record"


def round_to_100_percent(number_set, digit_after_decimal=0):
    """Convert a set of percentages
    which add up to 100% but no longer add up to 100 when rounded, to
    numbers which when rounded add up to 100. The excess/deficit is distributed
    using the algorithm described below.

    Args:
        number_set (sequence of X numbers)
        digit_after_decimal (int): how many decimals to round to, e.g.
            0 means the function will return integers.

    Returns:
        list: X numbers rounded according to digit_after_decimal

    From https://stackoverflow.com/q/25271388:

        This function take a list of number and return a list of percentage, which
        represents the portion of each number in sum of all numbers
        Moreover, those percentages are adding up to 100%!!!
        Notice: the algorithm we are using here is 'Largest Remainder'
        The down-side is that the results won't be accurate, but they are never
        accurate anyway:)

    """
    if np.sum(number_set) == 0:
        unround_numbers = [0 for x in number_set]
    else:
        unround_numbers = [
            x / float(np.sum(number_set)) * 100 * 10**digit_after_decimal
            for x in number_set
        ]
    decimal_part_with_index = sorted(
        [(index, unround_numbers[index] % 1) for index in range(len(unround_numbers))],
        key=lambda y: y[1],
        reverse=True,
    )
    remainder = 100 * 10**digit_after_decimal - np.sum(
        [int(x) if not np.isnan(x) else 0 for x in unround_numbers]
    )
    index = 0
    while remainder > 0:
        unround_numbers[decimal_part_with_index[index][0]] += 1
        remainder -= 1
        index = (index + 1) % len(number_set)
    return [
        (int(x) if not np.isnan(x) else np.nan) / float(10**digit_after_decimal)
        for x in unround_numbers
    ]


"""Reference lists for the BoM percentile classifications."""
bom_classes = list(percentile_to_bom_class([0, 5, 20, 50, 80, 95, 100]))
bom_classes_index = [bom_classes.index(x) for x in bom_classes]


def chunk(l, n=1000):
    """Yield successive n-sized chunks from a list l.

    Args:
        l (sequence)
        n (int)

    Returns:
        iterator: each iteration yields a list of length <= n

    .. code-block:: python

        >>> from dew_gwdata.utils import chunk
        >>> for x in chunk([0, 1, 2, 3, 4], n=2):
        ...     print(x)
        [0, 1]
        [2, 3]
        [4]

    """
    y = 0
    for i in range(0, len(l), n):
        y += 1
        yield l[i : i + n]
    if y == 0:
        yield l


class add_import_path:
    """Context manager for adding a path to sys.path temporarily."""

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        sys.path.insert(0, self.path)

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            sys.path.remove(self.path)
        except ValueError:
            pass


def df_to_shp(
    df,
    filename,
    x_col="longitude",
    y_col="latitude",
    convert_dt_cols=True,
    from_crs="epsg:7844",
    to_crs="epsg:8059",
    **kwargs,
):
    """Save a DataFrame as a shapefile.

    Args:
        df (pd.DataFrame): must have X and Y coordinates as columns
        filename (str): shapefile filename to create
        x_col (str): column of df with x coordinates
        y_col (str): column of df with y coordinates
        convert_dt_cols (bool): convert datetimes to string before
            writing, otherwise they are converted to epoch timestamps
            in the shapefile (not very useful!)
        from_crs (str): coordinate  system in df
        to_crs (str): coordinate system desired for shapefile

    Returns:
        nothing

    Other kwargs are passed to the geopandas.GeoDataFrame constructor.

    """
    if "crs" in kwargs:
        logger.warning('Please use the kwargs "from_crs" and "to_crs" instead of "crs"')
    if len(df):
        gdf = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df[x_col], df[y_col]),
            crs=from_crs,
            **kwargs,
        )
        gdf = gdf.to_crs(to_crs)
        if convert_dt_cols:
            for col in gdf.columns:
                value = gdf[col].iloc[0]
                if isinstance(value, datetime):
                    gdf[col] = gdf[col].astype(str)
        gdf.to_file(
            filename,
            driver="ESRI Shapefile",
        )


@contextlib.contextmanager
def cd(newdir):
    """Context manager for changing directory temporarily."""
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def append_comment_to_dataframe_column(df, indexer, column, comment):
    """Append a comment to rows from a dataframe.

    Args:
        df (pd.DataFrame): will be modified in-place.
        indexer (boolean pd.Series): e.g. df.well_id == 'ANG010'
        column (str): column for comment
        comment (str): comment to append to any existing values.

    Returns:
        pd.DataFrame: df containing changes (the modification happens in place)

    """
    if not column in df:
        df.loc[indexer, column] = comment
    else:
        df.loc[indexer, column] = df[indexer][column] + "\n" + comment
        df[column] = [v.strip("\n") if v == str(v) else v for v in df[column]]
    return df


def highlight_fields(lines, highlight_method=None):
    """Highlight fields in strings.

    Args:
        lines (sequence): list of strings
        highlight_method (str): either a symbol or colour

    Returns:
        list: the strings submitted, with fields replaced

    So for example if highlight_method = "blue", then this will
    replace lines like "the cat is <|angry|> today" with
    "the cat is <span style='background: blue'>angry</span> today"

    """
    if not highlight_method:
        replacements = []
    elif highlight_method[0] in ("*", "_", "`"):
        replacements = [("<|", highlight_method), ("|>", highlight_method)]
    else:
        replacements = [
            ("<|", f"<span style='background: {highlight_method}'>"),
            ("|>", "</span>"),
        ]
    new_lines = []
    for line in lines:
        for repl_args in replacements:
            line = str(line).replace(*repl_args)
        new_lines.append(line)
    return new_lines


def df_for_sql(df, *args, **kwargs):
    for column in df.columns:
        if column.endswith("_date"):
            df[column] = pd.to_datetime(df[column])
    return df



import colorsys
from datetime import date, datetime, timedelta
import io
import re
import os
import logging
import textwrap
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib import lines as mlines
from matplotlib import gridspec
from matplotlib import dates as mdates
from matplotlib import ticker as mticker
from matplotlib import colors as mcolors
import numpy as np
import pandas as pd
from adjustText import adjust_text
from scipy import stats
from PIL import Image, ImageChops
import pyproj

from .gwutils import *
from .utils import *


logger = get_logger()


def apply_technote_chart_style():
    # plt.style.use("ggplot")
    # plt.rcParams["axes.facecolor"] = "white"
    # plt.rcParams["axes.edgecolor"] = "#555555"
    plt.rcParams["figure.dpi"] = 150
    # plt.rcParams["font.family"] = "Segoe UI"
    # plt.rcParams["hatch.linewidth"] = 0.5
    plt.rcParams["figure.max_open_warning"] = 100


def convert_colour_format(cf, from_fmt, to_fmt, **kwargs):
    if from_fmt.startswith("rgb255"):
        if to_fmt.startswith("rgb1"):
            return rgb255_to_rgb1a(*cf, **kwargs)
        elif to_fmt.startswith("rgb255"):
            return rgb255_to_rgb255a(*cf, **kwargs)
        elif to_fmt == "hex":
            cf2 = rgb255_to_rgb1a(*cf, **kwargs)
            return rgba1_to_hex(*cf2, **kwargs)
    elif from_fmt.startswith("rgb1"):
        if to_fmt.startswith("rgb255"):
            return rgb1_to_rgb255a(*cf, **kwargs)
        elif to_fmt.startswith("rgb1"):
            return rgb1_to_rgb1a(*cf, **kwargs)
        elif to_fmt == "hex":
            return rgba1_to_hex(*cf, **kwargs)


def rgb255_to_rgb1a(r, g, b, a=255, auto_convert_alpha=True):
    """Convert RGBA values from 0 to 255.

    Args:
        r (float): red fraction from 0 to 255
        g (float): green fraction from 0 to 255
        b (float): blue fraction from 0 to 255
        a (float): alpha % from 0 to 255 or 0 to 1
        auto_convert_alpha (bool): if a <= 0 convert that fraction between 0 and 1
            into one between 0 and 255.

    Returns: r, g, b, a (tuple of floats between 0 and 1)

    """
    if a < 1:
        a *= 255
    elif a == 1:
        a = 255
    return (r / 255, g / 255, b / 255, a / 255)


def rgb1_to_rgb255a(r, g, b, a=1, alpha_as=255):
    """Convert RGBA values from 0-1 to 0-255.

    Args:
        r (float): red fraction from 0 to 1
        g (float): green fraction from 0 to 1
        b (float): blue fraction from 0 to 1
        a (float): alpha % from 0 to 255 or 0 to 1
        alpha_as (int): if 255, return alpha between 0 and 255; if 1, return alpha
            between 0 and 1

    Returns: r, g, b, a (tuple of floats between 0 and 255)

    """
    return (r * 255, g * 255, b * 255, a * alpha_as)


def rgb1_to_rgb1a(r, g, b, a=1):
    """Ensure an alpha value.

    Args:
        r (float): red fraction from 0 to 1
        g (float): green fraction from 0 to 1
        b (float): blue fraction from 0 to 1
        a (float, optional): alpha % from 0 to 1

    Returns: r, g, b, a (tuple of floats between 0 and 1)

    """
    return r, g, b, a


def rgb255_to_rgb255a(r, g, b, a=255):
    """Ensure an alpha value.

    Args:
        r (float): red fraction from 0 to 255
        g (float): green fraction from 0 to 255
        b (float): blue fraction from 0 to 255
        a (float, optional): alpha % from 0 to 255

    Returns: r, g, b, a (tuple of floats between 0 and 255)

    """
    return r, g, b, a


def rgba1_to_hex(r, g, b, a=1):
    """Convert RGBA (0-1) to a hexadecimal code.

    Args:
        r (float): red fraction from 0 to 1
        g (float): green fraction from 0 to 1
        b (float): blue fraction from 0 to 1
        a (float, optional): alpha % from 0 to 1

    Returns: string of hex code beginning with "#"

    """
    h = lambda v: hex(int(v * 255))[-2:]
    return f"#{h(r)}{h(g)}{h(b)}"


status_change_colours = {
    "Rising": (0, 176, 240),
    "Stable": (0, 176, 80),
    #     "Declining": (255, 255, 0), # original ArcMAP GSR colour (yellow)
    "Declining": (254, 178, 76),
    "Decreasing": (0, 176, 240),  # Salinity
    "Increasing": (254, 178, 76),  # Salinity
}
"""RGB (0 to 255) tuples for each of the possible values that "status_change" can take.

"""

wl_status_change_colours = [
    rgb255_to_rgb1a(*status_change_colours[c]) for c in wl_status_changes
]
"""RGB (0 to 255) tuples for each of the possible values that "status_change" can take
for WL data.

"""

tds_status_change_colours = [
    rgb255_to_rgb1a(*status_change_colours[c]) for c in tds_status_changes
]
"""RGB (0 to 255) tuples for each of the possible values that "status_change" can take
for TDS data.

"""


EXTRACT_METHOD_LUT = {
    "AIRL": "Air Lift",
    "BAIL": "Bailer",
    "BUCK": "Bucket",
    "EST": "Estimated",
    "FLOW": "Flow",
    "HAND": "Hand",
    "PUMP": "Pump",
    "UKN": "Unknown",
    "WMLL": "Windmill",
    "GRAB": "Grab",
    "SNDE": "Sonde",
    "?": "?",
}
"""Explanatory description for the "extract_method" values for TDS data.

"""

measured_during_lut = {
    "A": "Aquifer Test",
    "D": "Drilling",
    "F": "Field Survey",
    "S": "Final Sample on drilling completion",
    "G": "Geophysical Logging",
    "L": "Landowner Sample",
    "M": "Monitoring",
    "R": "Rehabilitation",
    "U": "Unknown",
    "W": "Well Yield",
    "?": "?",
}
"""Explanatory description for the "measured_during" values for TDS data.

"""

extract_method_markers = {
    "AIRL": "$a$",
    "BAIL": "v",
    "BUCK": ">",
    "EST": "$?$",
    "FLOW": "P",
    "HAND": "<",
    "PUMP": "o",
    "UKN": ".",
    "WMLL": "o",
    "GRAB": "<",
    "SNDE": "*",
    "?": ".",
}
"""matplotlib marker shape codes for the "extract_method" values for TDS data.

"""

meas_during_colours = {
    "M": "tab:cyan",  # 'Monitoring',
    "F": "tab:blue",  # 'Field Survey',
    "A": "tab:green",  # 'Aquifer Test',
    "W": "tab:olive",  # 'Well Yield',
    "L": "tab:pink",  # 'Landowner Sample',
    "G": "tab:purple",  # 'Geophysical Logging',
    "D": "tab:red",  # 'Drilling',
    "S": "tab:orange",  # 'Final Sample on drilling completion',
    "R": "tab:brown",  # 'Rehabilitation',
    "U": "tab:gray",  # 'Unknown',
    "?": "black",  # '?'
}
"""matplotlib string colours for the "measured_during" values for TDS data.

"""

removal_reason_colours = [
    "rosybrown",
    "coral",
    "peru",
    "orange",
    "darkgoldenrod",
    "tan",
    "darkkhaki",
    "olive",
    "yellowgreen",
    "green",
    "darkslategray",
    "darkcyan",
    "lightslategrey",
    "darkblue",
    "mediumorchid",
]
"""Sequence of matplotlib string colours for multiple removal reasons for data validation
figures.

"""


season_colours = {
    "0-summer": "sandybrown",
    "summer": "sandybrown",
    "1-autumn": "khaki",
    "autumn": "khaki",
    "2-winter": "deepskyblue",
    "winter": "deepskyblue",
    "3-spring": "yellowgreen",
    "spring": "yellowgreen",
}
"""Matplotlib string colours for the seasons.

"""


rainfall_colours = {"rainfall": (0 / 255, 176 / 255, 240 / 255, 1), "mean": "#bebebe"}
"""RGBA (0 to 1) tuple for the rainfall data figures, and a hex colour for the average
rainfall line.

"""


def adjust_lightness(color, amount=0.5):
    """Adjust colour to be lighter or darker.

    https://stackoverflow.com/a/49601444

    """
    import matplotlib.colors as mc
    import colorsys

    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def wrap(s, width=20, chars=("=", "-")):
    """Wrap a string."""
    if s is None:
        s = ""
    for char in chars:
        for prefix in (" ", ""):
            for suffix in ("", " "):
                s = s.replace(f"{prefix}{char}{suffix}", char)
    return "\n".join(textwrap.wrap(s, width=width))


class MonthlyColormap:
    """Define a circular colour mapping for the months of the year.

    Args:
        cmap (matplotlib Colormap, optional): suggested to use a
            circular one.
        rot (int): integer between 0 and 12 to rotate it by so that
            the seasons make sense

    Attributes:
        cmap (:class:`matplotlib.colors.ListedColormap`): the name of the
            colormap is "monthly_cmap"
        norm (:class:`matplotlib.colors.Normalize`)
        mappable (:class:`matplotlib.pyplot.cm.ScalarMappable`)

    """

    def __init__(self, cmap=None, rot=None):
        if cmap is None:
            cmap = plt.cm.get_cmap("twilight")
            rot = 7
        if rot is None:
            rot = 0
        colours = [cmap(((x + rot) / 12) % 1) for x in range(12)]
        self.cmap = mcolors.ListedColormap(colours, name="monthly_cmap")
        self.norm = mcolors.Normalize(1, 12)
        self.mappable = plt.cm.ScalarMappable(norm=self.norm, cmap=self.cmap)
        self.mappable.set_array([])


class BoMClassesColormap:
    """Define a colour mapping for the BoM percentile classes.

        .. figure:: figures/BoM_deciles.png

    Args:
        wrap_labels (bool): wrap the BoM descriptions on "much"

    Attributes:
        cmap (:class:`matplotlib.colors.ListedColormap`): the name of the
            colormap is "monthly_cmap"
        norm (:class:`matplotlib.colors.Normalize`)
        mappable (:class:`matplotlib.pyplot.cm.ScalarMappable`)

    """

    # This is from the State of the Climate 2018 report:
    # colours = [
    #     (239, 84, 45),    # 0: Lowest on record
    #     (255, 193, 85),   # 1: Very much below average
    #     (255, 233, 170),  # 2: Below average
    #     (255, 255, 255),  # 3: Average
    #     (158, 225, 223),  # 4: Above average
    #     (87, 208, 221),   # 5: Very much above average
    #     (11, 118, 169),   # 6: Highest on record
    # ]

    #: RGB (0 to 255) tuples for the seven class codes:
    colours = [
        (178, 24, 43),  # 0: Lowest on record
        (239, 138, 98),  # 1: Very much below average
        (253, 219, 199),  # 2: Below average
        (247, 247, 247),  # 3: Average
        (209, 229, 240),  # 4: Above average
        (103, 169, 207),  # 5: Very much above average
        (33, 102, 172),  # 6: Highest on record
    ]

    #: RGB (0 to 255) tuples for eight class codes (i.e. index 0 -> code "no data").
    colours_nodata = [
        (120, 120, 120),
    ] + colours  # -1: not applicable

    #: The names of each of the BoM classifications
    class_names = [
        "Lowest on record",
        "Very much below average",
        "Below average",
        "Average",
        "Above average",
        "Very much above average",
        "Highest on record",
    ]

    #: foreground colours for each BoM class name
    foreground_colours = ["white", "white", "white", "black", "black", "white", "white"]

    #: foreground colours for each BoM class name (version 2... ?)
    foreground_colours_2 = [
        "white",
        "black",
        "black",
        "black",
        "black",
        "black",
        "white",
    ]

    @property
    def hexdict(self):
        """Dictionary of hex colour codes for each class name."""
        return {c: rgba1_to_hex(*self.class_to_rgba(c)) for c in self.class_names}

    def __init__(self, wrap_labels=True):
        self.cmap = mcolors.ListedColormap(self.colours_1, name="bom_cmap")
        self.norm = mcolors.Normalize(0, 6)
        self.mappable = plt.cm.ScalarMappable(norm=self.norm, cmap=self.cmap)
        self.mappable.set_array([])
        self.wrap_labels = wrap_labels

        #: A version of the colourmap including a "nodata" value.
        self.cmap_nodata = mcolors.ListedColormap(
            self.colours_nodata_1, name="bom_cmap2"
        )
        #: A matplotlib Normalize instance of the colourmap including a "nodata" value.
        self.norm_nodata = mcolors.Normalize(0, 7)
        #: A matplotlib ScalarMappable for the colourmap including a "nodata" value.
        self.mappable_nodata = plt.cm.ScalarMappable(
            norm=self.norm_nodata, cmap=self.cmap_nodata
        )
        self.mappable_nodata.set_array([])
        self.wrap_labels = wrap_labels

    @property
    def colours_1(self):
        """List of RGB (0 to 1) tuples for each BoM class."""
        return [tuple([x / 255 for x in rgb]) for rgb in self.colours]

    @property
    def colours_nodata_1(self):
        """List of RGB (0 to 1) tuples for each BoM class (including nodata as 0)."""
        return [tuple([x / 255 for x in rgb]) for rgb in self.colours_nodata]

    def colours_rgba(self, alpha=1):
        """Add alpha to the RGB colours.

        Args:
            alpha (float): between 0 and 1, opacity (1 == opaque, 0 = transparent)

        Returns: list of (r, g, b, a) tuples for all the classes (rgba between 0 and 1)

        """
        return [tuple([x / 255 for x in rgb] + [alpha]) for rgb in self.colours]

    def class_to_rgba(self, bom_class, alpha=1):
        """Transform a BoM class name e.g. "Lowest on record" to an RGBA tuple.

        Args:
            bom_class (str): one of the BoM class definitions
            alpha (float): between 0 (transparent) and 1 (opaque)

        Returns: (r, g, b, a) tuple (between zero and one)

        """
        if bom_class == "No data":
            return self.colours_nodata_1[0]
        idx = self.class_names.index(bom_class)
        return self.colours_rgba(alpha=alpha)[idx]

    def class_to_foreground_colour(self, bom_class):
        """Transform a BoM class name e.g. "Lowest on record" to a foreground colour.

        Args:
            bom_class (str): one of the BoM class definitions

        Returns: str (matplotlib colour)

        """
        idx = self.class_names.index(bom_class)
        return self.foreground_colours[idx]

    def class_to_foreground_colour_2(self, bom_class):
        """Transform a BoM class name e.g. "Lowest on record" to a foreground colour.

        Args:
            bom_class (str): one of the BoM class definitions

        Returns: str (matplotlib colour)

        """
        idx = self.class_names.index(bom_class)
        return self.foreground_colours_2[idx]

    @property
    def labels(self):
        """Wrap labels (if required) on the word 'much' for the class names."""
        labels = [
            "Lowest on record",
            "Very much below average",
            "Below average",
            "Average",
            "Above average",
            "Very much above average",
            "Highest on record",
        ]
        if self.wrap_labels:
            return [l.replace("much ", "much\n") for l in labels]
        else:
            return labels

    @property
    def labels_nodata(self):
        """Wrap labels (if required) on the word 'much' for the class names (including
        nodata value).

        """
        labels = [
            "No data",
            "Lowest on record",
            "Very much below average",
            "Below average",
            "Average",
            "Above average",
            "Very much above average",
            "Highest on record",
        ]
        if self.wrap_labels:
            return [l.replace("much ", "much\n") for l in labels]
        else:
            return labels

    def plot_legend(self, cax, cbar_labels=None, nodata=False, **kwargs):
        """Plot legend into an Axes.

        Args:
            cax (matplotlib Axes): axes to fill with a legend.

        Returns: matplotlib Colorbar object.

        """
        # fig = plt.figure()
        if nodata:
            cmap = self.cmap_nodata
        else:
            cmap = self.cmap
        cset = cax.scatter([], [], c=[], cmap=cmap)
        cbar = plt.colorbar(cset, cax=cax, **kwargs)
        self.fix_ticklabels(cbar, cbar_labels=cbar_labels, nodata=nodata)
        return cbar

    def fix_ticklabels(self, cbar, wrap_labels=None, cbar_labels=None, nodata=False):
        """Fix colorbar ticks and labels.

        Args:
            cbar (:class:`matplotlib.Colorbar`)
            wrap_labels (bool)

        """
        if wrap_labels is None:
            wrap_labels = self.wrap_labels
        c0, c1 = cbar.mappable.get_clim()
        cbar_ticks = []

        if nodata:
            cbar_divisions = 8
        else:
            cbar_divisions = 7
        cbar_step = (c1 - c0) / cbar_divisions
        if not cbar_labels:
            cbar_labels = []
            make_cbar_labels = True
        else:
            make_cbar_labels = False
        for i, value in enumerate(range(0, cbar_divisions)):
            if make_cbar_labels:
                cbar_labels.append(self.labels[i])
            cbar_ticks.append(((cbar_step * i) + c0) + (cbar_step / 2))
        cbar.set_ticks(cbar_ticks)
        cbar.set_ticklabels(cbar_labels)


def fix_ticklabels(cbar, cbar_labels=None):
    c0, c1 = cbar.mappable.get_clim()
    cbar_ticks = []
    if cbar_labels is None:
        cbar_labels = []
        make_labels = True
    else:
        make_labels = False
    cbar_divisions = len(cbar_labels)
    cbar_step = (c1 - c0) / cbar_divisions
    for i, value in enumerate(range(0, cbar_divisions)):
        if make_labels:
            cbar_labels.append(colours[i])
        cbar_ticks.append(((cbar_step * i) + c0) + (cbar_step / 2))
    cbar.set_ticks(cbar_ticks)
    cbar.set_ticklabels(cbar_labels)
    return cbar


def plot_colorbar(
    cax,
    colours,
    cbar_labels=None,
    split_cbar_labels_on="Very much",
    orientation="vertical",
):
    if split_cbar_labels_on:
        cbar_labels = [
            l.replace(split_cbar_labels_on, split_cbar_labels_on + "\n")
            for l in cbar_labels
        ]
    cmap = mcolors.ListedColormap(colours, name="custom_colours")
    cset = cax.scatter([], [], c=[], cmap=cmap)
    cbar = plt.colorbar(cset, cax=cax, orientation=orientation)
    cbar = fix_ticklabels(cbar, cbar_labels=cbar_labels)
    for axis in ["top", "bottom", "left", "right"]:
        cax.spines[axis].set_linewidth(0.2)
    return cbar


def trim_whitespace_from_image(fn, append=None, new_fn=None):
    fn = str(fn)
    if new_fn is None:
        new_fn = fn
    if append:
        root, ext = os.path.splitext(fn)
        new_fn = root + append + ext

    def trim(im):
        bg = Image.new("RGB", im.size, im.getpixel((0, 0)))
        logger.debug(f"bg image: {bg}")
        diff = ImageChops.difference(im.convert("RGB"), bg)
        logger.debug(f"ImageChops.difference(im, bg): {diff}")
        diff = ImageChops.add(diff, diff, 2.0, -100)
        bbox = diff.getbbox()
        logger.debug(f"bbox() = {bbox}")
        if bbox:
            logger.debug(f"found bbox to trim: {bbox}")
            return im.crop(bbox)
        else:
            logger.debug("Warning! Did not find any whitespace margin to trim.")
            return im

    im = Image.open(fn)
    im = trim(im)
    im.save(new_fn)


def plot_internal_classes_for_aggregate_resources(dfs, class_names, colours, col):
    """Plot a summary stacked bar chart showing classes for aggregate resources.

    Args:
        dfs (dict): dict; keys are the aggregate resource keys (really they are
            just labels for the x axis on the chart); keys should be a pandas
            DataFrame in each case.
        class_names (list): names for each class e.g. the "Decreasing", "Stable",
            "Increasing". They should be ordered such as will appear on the chart
            from bottom to top.
        colours (list): RGBA tuple colours for each class.
        col (str): column for *dfs* to extract the class value. The number of rows
            with a given value will be used to determine the height of each
            stacked bar segment.

    Returns: matplotlib Axes

    Examples follow below for Adelaide Plains:

        >>> import wrap_technote as tn
        >>> report = tn.load_report("Adelaide Plains", "2019-20")

    To do current WL rankings:

        >>> wlranks = report.get_aggregate_resources_data(
        ...     "WL", "recovery_wl_data", "current ranked WLs"
        ... )
        >>> bcmap = tn.BoMClassesColormap()
        >>> tn.plot_internal_classes_for_aggregate_resources(
        ...     wlranks, bcmap.class_names, bcmap.colours_rgba(), col="rswl_bom_class"
        ... )

    Five-year WL trends:

        >>> wltrends = report.get_aggregate_resources_data(
        ...     "WL", "recovery_wl_trends", "final trends"
        ... )
        >>> plot_internal_classes_for_aggregate_resources(
        ...     wltrends, wl_status_changes, wl_status_change_colours, col="status_change"
        ... )

    Five-year TDS trends:

        >>> tdstrends = report.get_aggregate_resources_data(
        ...     "TDS", "salinity_trends", "final trends"
        ... )
        >>> plot_internal_classes_for_aggregate_resources(
        ...     tdstrends,
        ...     tds_status_changes,
        ...     tds_status_change_colours,
        ...     col="status_change",
        ... )

    """
    pct_counts = {}
    raw_counts = {}
    labels = []
    across_all_counts = {c: 0 for c in class_names}
    for name, df in dfs.items():
        counts = {c: 0 for c in class_names}
        counts.update(df[col].value_counts().to_dict())
        for key, value in counts.items():
            across_all_counts[key] += value
        total_wells = sum(counts.values())
        if total_wells == 0:
            total_wells = np.nan
        pct_counts[name] = [counts[c] / total_wells * 100 for c in class_names]
        pct_counts[name] = round_to_100_percent(pct_counts[name])
        raw_counts[name] = [counts[c] for c in class_names]
        labels.append(name)

    width = (len(labels) + 1) * 1.4
    width_ratio_factor = 2.0
    fig = plt.figure(figsize=(width, 3.6))
    gs = gridspec.GridSpec(1, 2, width_ratios=(width * width_ratio_factor, 1))
    ax = fig.add_subplot(gs[0])
    cax = fig.add_subplot(gs[1])
    width = 0.3
    ax.axhline(50, lw=0.5, dashes=(20, 5), color="k", alpha=0.2)
    pct_counts_arr = np.array(list(pct_counts.values()))
    pct_counts_arr_cum = pct_counts_arr.cumsum(axis=1)
    n_wells = sum(across_all_counts.values())

    xlabels = [l.replace("_", "\n") + f"\n({sum(raw_counts[l])})" for l in labels]
    xticks = np.arange(len(xlabels))
    texts = []
    bars = []
    label_counts = np.zeros_like(xticks)

    for i, class_name in enumerate(class_names):
        colour = colours[i]
        heights = pct_counts_arr[:, i]
        starts = pct_counts_arr_cum[:, i] - heights
        bar = ax.bar(
            xticks,
            heights,
            bottom=starts,
            color=colour,
            width=width,
            linewidth=0.5,
            edgecolor="#555555",
        )
        bars += bar
        for j, label in enumerate(labels):
            category = pct_counts_arr[j, i]
            top = pct_counts_arr_cum[j, i]
            if category > 0:
                label_counts[j] += 1
                offset = width / 2 + 0.03
                ha = "left"
                if label_counts[j] % 2:
                    offset *= -1
                    ha = "right"
                text = ax.text(
                    j + offset,
                    top - (category / 2),
                    f"{pct_counts[label][i]:.0f}%\n({raw_counts[label][i]})",
                    color=adjust_lightness(colour, 0.5),
                    fontsize="small",
                    ha=ha,
                    va="center",
                    ma="center",
                )
                texts.append(text)
    cbar = plot_colorbar(cax, colours, cbar_labels=class_names)
    cbar.ax.tick_params(labelsize="small")
    ax.set_ylabel("Percentage of wells (number of wells)", fontsize="small")
    ax.set_xticks(xticks)
    ax.set_xticklabels(xlabels)
    plt.setp(ax.get_xticklabels(), rotation=0, ha="center", fontsize="small")
    for spine in ["top", "right"]:
        ax.spines[spine].set_visible(False)
    ax.set_xlim(xticks[0] - 0.5, xticks[-1] + 0.5)
    ax.figure.tight_layout()
    return ax


def plot_trends_internal(
    trends, title, class_names, param_yaxis_label, param_col="slope_yr"
):
    """Plot bar chart of trend triclasses for internal use.

    Args:
        trends (DataFrame): e.g. from
            `resource.read_data("recovery_wl_trends", "final_trends")
        title (str): chart title e.g. resource_key

    Returns: matplotlib figure


    """

    class_colours = [rgb255_to_rgb1a(*status_change_colours[c]) for c in class_names]
    class_counts = {c: 0 for c in class_names}
    class_counts.update(trends.status_change.value_counts().to_dict())

    fig = plt.figure(figsize=(10, 4))
    gs = gridspec.GridSpec(1, 3, width_ratios=(10, 10, 1))
    ax = fig.add_subplot(gs[0])
    ax3 = fig.add_subplot(gs[1])
    ax2 = fig.add_subplot(gs[2])

    bc_range = [i for i in range(len(class_names))]
    bc_counts = [class_counts[x] for x in class_names]
    ax.barh(bc_range, bc_counts, color=class_colours, edgecolor="k", lw=0.8)
    ax.set_yticks(bc_range)
    _ = ax.set_yticklabels(class_names, fontsize="small")
    label_ypos = np.asarray(bc_counts) + (max(bc_counts) / 30)
    percentage_counts = [y / sum(bc_counts) * 100 for y in bc_counts]
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    bottom = 0
    for i in range(len(bc_range)):
        if bc_counts[i] > 0:
            ax.text(
                label_ypos[i],
                bc_range[i],
                f"{bc_counts[i]} wells ({percentage_labels[i]}%)",
                va="center",
                ha="left",
                color="#222222",
                fontsize="small",
            )
        ax2.bar(
            [1],
            [percentage_counts[i]],
            bottom=bottom,
            facecolor=class_colours[i],
            edgecolor="k",
            lw=0.5,
        )
        bottom += percentage_counts[i]
    ax.set_xlabel("Number of wells", fontsize="medium")
    y0, y1 = ax.get_xlim()
    ax.xaxis.set_major_locator(
        mticker.MaxNLocator(nbins="auto", steps=[1, 2, 5, 10], integer=True)
    )
    ax.set_xlim(0, y1 + np.ceil((max(bc_counts) / 30)))
    for sp in ["top", "right"]:
        ax.spines[sp].set_visible(False)
    for sp in ["top", "bottom", "left"]:
        ax2.spines[sp].set_visible(False)
    ax2.tick_params(
        axis="y",
        left=False,
        labelleft=False,
        right=True,
        labelright=True,
    )
    ax2.set_xticks([])
    ax2.set_xlim(0, 1.5)
    ax2.set_ylim(-0.5, 100)
    ax2.set_ylabel(
        f"Percentage of wells ({sum(bc_counts)} wells in total)", fontsize="small"
    )
    ax2.yaxis.set_label_position("right")
    median_class = get_median_trend_triclass(
        trends["status_change"], classes=class_names
    )
    median_slope = trends[param_col].median()
    ax2.annotate(
        f"Median well: {median_class} at {median_slope:.2f} {param_yaxis_label}",
        (1.4, 50),
        (0.4, 50),
        arrowprops=dict(arrowstyle="->", color="k"),
        color="k",
        rotation=90,
        ha="right",
        va="center",
        fontsize="small",
    )

    ax.set_title(title, fontsize="medium")

    ax3_class_names = [c for c in class_names if not c == "Stable"]
    xlabels = ["All\nwells"] + [f"{c}\nwells" for c in ax3_class_names]
    data_series = [trends[param_col]]
    for class_name in ax3_class_names:
        data_series.append(trends.loc[trends.status_change == class_name, param_col])
    data = ax3.boxplot(data_series)
    ax3.set_xticklabels(xlabels)
    ax3.set_ylabel(f"Trendline slope ({param_yaxis_label})")
    ax3.set_xlabel(
        "Box from 25 to 75th percentile\norange line = median", fontsize="small"
    )
    ax3.axhline(0, lw=0.5, color="b")

    fig.tight_layout()
    return fig
